#include "config.h" //konfigurace

#ifdef use_ode_lib
#include <ode/ode.h>
 #ifdef _WIN32
 #pragma comment( lib, "ode.lib" )
 #pragma comment( lib, "opcode.lib" )
 #endif
#endif
#include "build.h"
#include "bilbo.h"
#include "b_config.h"
#include "b_types.h"
#include "b_mem.h"
#include "b_string.h"
#include "b_util.h"
#include "b_math.h"
#include "b_gl.h"
#include "b_fs.h"
#include "b_model3d.h"
#include "b_tex.h"
#include "b_con.h"
#include "b_world.h"
#include "b_sound.h"
#include "b_ai.h"
#include "b_map.h"
#include "b_coll.h"
#include "b_menu.h"
#include "b_gener.h"
#include "b_net.h"

//TODO: add "nosound" variable

#define version "0.3.1 build " BUILDNUMBER " (" BUILDTIME ")" //verze enginu

#define introtext1 "Scorch Blaster 3D version " version //1.radka textu co se objevi pri spusteni hry
#define introtext2 "(c) 2002,2003 by Martin 'Bilbo' Petricek" //2.radka
#define introtext3 "Email: <petricm@seznam.cz>" //3.radka
#define introtext4 "WWW: http://www.volny.cz/petricm" //4.radka
//TODO: add entry splashscreen with infotext & logo

 int game_paused;                   //1 pokud je pauznuta hra

 //consts
 bfloat fog_distance=3000;          // hustota mlhy, cim vyssi cislo, tim ridsi
 vector fog_color={0.5,0.5,1.0};    // barva mlhy
 #define apptitle "Scorch Blaster 3D"     // titulek okna

 bfloat bomb_throwup_angle=0;        //uhel, pod kterym jsou vyhazovany bomby (oproti uhlu pohledu)
 bfloat bomb_throw_speed=70;        //rychlost, se kterou jsou bomby hazeny

 #define turn_speed (-0.18f/100.0f)   // rychlost otaceni
 #define roll_speed (0.18f/100.0f)    //rychlost otaceni kolem vlastni osy
 bfloat h_sensitivity=0.01;        // citlivost mysi - horizontalni
 bfloat v_sensitivity=0.01;        // citlivost mysi - vertikalni
 bfloat sensitivity=1.0;	          //citlivost mysi - nasobek
 #define person_width 0.3f        // bounding box - sirka postavy
 #define person_height_up 0.1f    // bounding box - vyska postavy od oci nahoru
 #define person_height_down 0.5f  // bounding box - vyska postavy od oci dolu
 bfloat vgamma=1.0;             //gamma korekce
 int use_skybox=1;              //zdali je pouzit skybox
 int draw_fps=1;		// zobrazovani fps (0=off, 1=fps, 2=fps+poly, 3=detailed)
 int phase,z_trick=0;           //figl na urychleni renderingu - trochu ale blbne, defaultne disablovan
 gl4fcolor zero4color={0.0,0.0,0.0,0.0};	//nulovy vektor (4*GLfloat)
 gl4f zero_4={0.0,0.0,0.0,0.0};	//nulovy vektor (4*bfloat)

 //byte a[si1][si2][si3];
 int mousex,mousey; //souradnice mysiho kurzoru v okne
 int omousex,omousey; //souradnice mysiho kurzoru v okne minuly frame
 int freemouse=0;//1 pokud mys muze opustit okno (potom nefunguje korektne mouselook), 0 pokud nemuze
 int fullscreen=0;//1 pokud hra bezi fullscreen, 0 pokud v okne
// bbox debug_box;
// aa_bbox debug_abox;
 int window_minimized=0;//1 pokud je okno prave minimalizovane (a tudiz se na nej nemusi kreslit)

int fraglimit=10;//omezeni na pocet fragu - po nahrani tolika fragu hra konci
int survival=0;//survival mode

typedef struct {  //definice zbrane
 char *classname;     //classname objektu, ktery je pri vystrelu vypusten (projektil)
 char *name;          //jmeno zbrane (zobrazeno uzivateli)
 int fire_delay;      //pauze mezi vystrely
 int switch_off_delay;//doba prepnuti na tuto zbran (doba nez se zbran vytahne)
 int switch_on_delay; //doba prepnuti z teto zbrane (doba nez se zbran schova)
} weapon_def;

//TODO: weapon database v souboru
#define num_weapons 2 //pocet zbrani
weapon_def weapdef[num_weapons]={
 {"bomb",  "Bomb",       600,120,120},
 {"bomb_s","Sticky bomb",300,120,120},
};

/** vola se pri zmene velijkosti obrazovky (nebo velikosti fontu, nebo ...) */
void b3d_recalc_window(void) {
 con_width=(int)(maxx/fontsizex);
}

/** callback funkce volana pri zmene rozmeru okna */
cdecl_void b3d_callbackreshape(int x,int y) {
 maxx=x;
 maxy=y;
 b3d_recalc_window();
 player_realloc_screens();
 glutPostRedisplay();
}

#define vidmode "640x480:32@60" //defaultni videomod u fullscreen zobrazeni

/** inicializace OpenGL, vytvoreni okna ... */
void b3d_init(int w,int h,int *argc, char** argv) {
 if (*argc>1 && eqi(argv[1],"fs")) fullscreen=1;
// glerr("b3d_init:begin");
 glutInit(argc,argv);
// glerr("b3d_init:after glutinit");
 glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
// glerr("b3d_init:after init 1");//tady to jeste chyby nevraci, pac neexistuje opengl context
 if (fullscreen) {
  glutGameModeString(vidmode);
  glutEnterGameMode();
 } else {
  glutInitWindowSize(w,h);
  glutInitWindowPosition(0,0);
  b3d_win=glutCreateWindow(apptitle);
 }
 glerr("b3d_init:after init 2");
 //glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 glerr("b3d_init:after blend");
 gle_tree=glext_parse((const char*)glGetString(GL_EXTENSIONS));
 if (gl_ext("GL_NV_fog_distance")) {
  glFogi(FOG_DISTANCE_MODE_NV,EYE_RADIAL_NV); //lepe vypadajici mlha
  //TODO: kdyz to neni, tak zkusit jeste GL_EXT_fog_coord (coz ale chce vice prace)
 }
 glerr("b3d_init:after extensions");
 maxx=w;
 maxy=h;
 b3d_recalc_window();
 glerr("b3d_init:end");
}

/** vycentruje mys - presune pointer doprostred okna/obrazovky */
void b3d_centermouse(void) {
#ifndef dedicated_server
 mousex=maxx/2;
 mousey=maxy/2;
 omousex=mousex;
 omousey=mousex;
 glutWarpPointer(mousex,mousey);
#endif
}

#define clip360(x) if((x)<2*pi)(x)+=2*pi;if((x)>2*pi)(x)-=2*pi; //osekne uhel na -360..360, neboli -2pi..2pi .. aby to neztracelo precision

/** callback funkce na pohyb mysi (mouselook, atd ...) */
void b_mousephysics(void) {
 if (mouseplayer && !game_paused) {//pokud nekdo ovlada mys
  mouseplayer->data.human.r_angle+=(mousex-omousex)*h_sensitivity*sensitivity;
  mouseplayer->data.human.r_angle2-=(mousey-omousey)*v_sensitivity*sensitivity;
  if (groundcontrol) { clip360(mouseplayer->data.human.r_angle);}
  if (mouseplayer->data.human.r_angle2>pi/2) mouseplayer->data.human.r_angle2=pi/2;
  if (mouseplayer->data.human.r_angle2<-pi/2) mouseplayer->data.human.r_angle2=-pi/2;
 }
 if (!freemouse) {
  if ((mousex==omousex) && (mousey==omousey)) return;
  b3d_centermouse();
 }
}

/** handler na pohyb mysi */
cdecl_void b_mouse(int x,int y) {
 omousex=mousex;
 omousey=mousey;
 mousex=x;
 mousey=y;
 b_mousephysics();
}

#ifndef dedicated_server

/** vlozi do sceny 1 svetlo*/
void b3d_light(int num,const gl4fcolor *position,const gl4fcolor *color) {
 glerr("b3d_light:begin");
 glLightfv(GL_LIGHT0+num,GL_POSITION,(float *)position);
 glLightfv(GL_LIGHT0+num,GL_AMBIENT, (float *)zero4color);
 glLightfv(GL_LIGHT0+num,GL_DIFFUSE, (float *)color);
 glLightfv(GL_LIGHT0+num,GL_SPECULAR,(float *)zero4color);
 glEnable(GL_LIGHT0+num);
 glerr("b3d_light:end");
}

/** zapne osvetleni sceny (plusminus zeshora)*/
void b3d_light_init() {
 GLfloat pos[4]={0.5,0.5,0,0};
 GLfloat one[4]={1,1,1,1};
 GLfloat amb[4]={0.1,0.1,0.1,0.1};
 glerr("b3d_light_init:begin");
 glLightfv(GL_LIGHT0,GL_POSITION,pos);
 glLightfv(GL_LIGHT0,GL_AMBIENT, one);
 glLightfv(GL_LIGHT0,GL_DIFFUSE, amb);
 glLightfv(GL_LIGHT0,GL_SPECULAR,amb);
 glEnable(GL_LIGHT0);
 glerr("b3d_light_init:end");
}

#endif
/** inicializace OpenGl na zacatku framu (vycisteni obrazovky, depth bufferu, atd ....)*/
void b3d_frame_init(const vector fogcolor) {
#ifndef dedicated_server
 glerr("b3d_frame_init:begin");
 glClearColor(fogcolor.x*vgamma,fogcolor.y*vgamma,fogcolor.z*vgamma,0.0);
 glEnable(GL_DEPTH_TEST);
 if (z_trick) { //maly trik - neni nutno mazat depth buffer, cimz se setri cas.
  //Nevyhoda je, ze snizuje presnost depth bufferu o 1 bit (z 32 bitu na 31, z 16 bitu na 15)
  if (!(use_skybox && current_skymap)) glClear(GL_COLOR_BUFFER_BIT);
  if (phase) {
   glDepthRange(0,0.5);
   glDepthFunc(GL_LESS);
  } else {
   glDepthRange(1,0.5);
   glDepthFunc(GL_GREATER);
  }
  phase=!phase;
 } else {
  glDepthRange(0,1);
  if (use_skybox && current_skymap) glClear(GL_DEPTH_BUFFER_BIT);
  else glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_LESS);
 }
 b3d_light_init();//zapne osvetleni
 glerr("b3d_frame_init:end");
#endif
}

/** inicializace OpenGl na zacatku framu pro kazdou obrazovku ve split screenu */
void b3d_frame_single_init(bfloat range) {
#ifndef dedicated_server
 gl4fcolor fogcolor2;
 int i;
 glerr("b3d_frame_single_init:begin");
 if (range<0.01) {
  glDisable(GL_FOG);
  return;
 }
 glEnable(GL_FOG);
 glFogi(GL_FOG_MODE,GL_LINEAR);
// glHint(GL_FOG_HINT,GL_FASTEST);
 glFogf(GL_FOG_START,0.0);
 glFogf(GL_FOG_END,range);
 for (i=0;i<3;i++) fogcolor2[i]=((vector2*)&fog_color)->a[i]*vgamma;
 fogcolor2[3]=0;
 glFogfv(GL_FOG_COLOR,fogcolor2);
 glerr("b3d_frame_single_init:end");
#endif
}

/** zmeni behem hry pouzitou skymapu*/
void cmd_change_skymap(const char *param) {
#ifndef dedicated_server
 fs_loadskymap(param);
#endif
 if (!param || !*param) con_printfv("Unloaded skymap");
 else con_printfv("Loaded skymap %s",param);
}


/** vytvori modely s explozemi */
void b_create_explosions() {
 tmpmodel *m;
 m=model_begin();
 model_face_begin(m,255,255,255,255,"exp_e");
 model_vertex(m,0.7,0.7,0,0,0.7);
 model_face_end(m);
 model_face_begin(m,255,255,255,255,"exp_e");
 model_vertex(m,1,1,0,0,0);
 model_face_end(m);
 model_face_begin(m,255,255,255,255,"exp_e");
 model_vertex(m,0.7,0.7,0,0,-0.7);
 model_face_end(m);
 explosion_model=model_end(m);
 model_free(m);
 //TODO: tu neznicitelnost modelu bychom meli delat jinak .. byt takhle to zatim funguje
 explosion_model->manager=malloc0(sizeof(model));		        
 explosion_model->manager->mem=explosion_model;
 explosion_model->manager->refcount=INT_MAX;//stejne ho to ignoruje
}

/** nahraje zakladni textury (font, pozadi konzole....) a jina data */
void b_loaddata(void) {
 //textury se nahravaji "on demand" pomoci texture manageru, zde se nahrava jen par fixnich (fonty a tak ...)
#ifndef dedicated_server
 tex_console=b_loadtexture_r("console");
 tex_menu[0]=b_loadtexture_r("menu");
 tex_menu[1]=b_loadtexture_r("menutab");
 tex_menu[2]=b_loadtexture_r("menutab_p");
 tex_menu[3]=b_loadtexture_r("menubut");
 tex_font=b_loadtexture_r("font/ms_12x16");
//CRITICAL TODO: na nasledujicim prikazu to zatuhne nekde v NVOPENGL.DLL, obcas v KERNEL32.DLL
// fs_loadskymap("_sk");//TODO: kdyz to tu je , neboli kdyz se pri startu skymap nahrava dvakrat, tak to blbne 
 b_create_explosions();
#endif
}

//TODO: add camera-eye-offest (left/right eye) pro TRUE-3d zobrazeni

/** nahraje mapu z disku, ze spravneho adresare*/
worldspawn *fs_loadworld(const char* param) {
 //TODO: loadnout nejdriv .b4d potom .map
 char map[257];
 map[256]=0;
 _snprintf(map,sizeof(map)-1,"%s/%s.b4d",mapdir,param);
 return loadworld(map);
}

/** prikaz na zmenu mapy. Bude i s priponou */
void cmd_map(const char* param) {
#ifdef map_named_model
 int oldserial=model_serial;
#endif
 worldspawn *w;
 bench_start();
#ifdef map_named_model
 model_rename_temporaries(':','|');
#endif
 w=fs_loadmap(param,1);
 if (w) {
  if (world) {
   world_stop(world);
   freeworld(world);
#ifdef map_named_model
   model_free_temporaries('|');//stare tmp modely pryc
#endif
  }
  world=w;
  world_start(world);
  con_printfv("Map %s loaded",param);
  map_change_handler();//v sitove hre notifikopvat klienty at si taky zmeni mapu
 } else {
#ifdef map_named_model
  model_rename_temporaries('|',':');
  model_serial=oldserial;
#endif
  con_printf("Map %s: load error",param);
 }
 bench_result();
}

/** true pokud je textura v texture pameti */
GLboolean tex_isresident(texture *t) {
 GLboolean i;
 glAreTexturesResident(1,&t->id,&i);
 return i;
}

/** callback funkce pro vypis textur */
void _cmd_texlist(char *name,texture *t) {
 con_printf("%s [%2d]%s %3dx%3dx%2d (R%d)",name,t->refcount,(t->mem==failsafetex->mem)?" MISSING":"",t->mem->head.x,t->mem->head.y,t->mem->head.bpp,tex_isresident(t));
}

/** prikaz na vypis nahranych textur (debugovaci) */
void cmd_texlist(const char* param) {
 con_printf("Texture list:");
 tree_foreach(tex_tree,(tree_func *)_cmd_texlist);
 con_printf("Texture memory:%d",tex_memory);
 unused(param);
}

/** likvidace starych textur z pameti */
void cmd_flushtex(const char *param) {
 int texmem=tex_memory;
 tex_flush();
 con_printf("Textures flushed (from %d bytes to %d bytes)",texmem,tex_memory);
 unused(param);
}

size_t mem_usage;//memory usage for modellist,classlist ...
int item_count;//item count for modellist,classlist ...

#define modelsize_mem(m) sizeof(model3d)+(m)->nface*sizeof(modelface)+(m)->nfacev*sizeof(facecoord)+(m)->nvert*sizeof(vertex) //velikost modelu v pameti

/** callback funkce pro vypis modelu */
void _cmd_modellist(char *name,model *m) {
 int mo=sizeof(model)+strlen(m->name)+1+modelsize_mem(m->mem);
 con_printf("%s [%d] vert=%d face=%d (%g %g %g,%g %g %g) (%d b)",name,m->mem->manager?m->mem->manager->refcount:-999999,m->mem->nvert,m->mem->nface,all3d(m->mem->abox.lo),all3d(m->mem->abox.hi),mo);
 item_count++;
 mem_usage+=mo;
}

/** prikaz na vypis nahranych modelu (debugovaci) */
void cmd_modellist(const char* param) {
 int mo;
 con_printf("Model list:");
 mem_usage=0;item_count=0;
 tree_foreach(model_tree,(tree_func *)_cmd_modellist);
 //worldspawn
 mo=sizeof(model)+modelsize_mem(world->m);
 con_printf("<worldspawn> [--] vert=%d face=%d (%g %g %g,%g %g %g) (%d b)",world->m->nvert,world->m->nface,all3d(world->m->abox.lo),all3d(world->m->abox.hi),mo);
 item_count++;
 mem_usage+=mo;
 //results
 con_printf("%d bytes, %d items",mem_usage,item_count);
 unused(param);
}

/** debugovaci funkce na zjisteni kolik mista zabiraji v pameti properties */
size_t prop_size(properties *p) {
 if (!p) return 0;
 return prop_size(p->next)+strlen(p->k)+strlen(p->v)+2+sizeof(properties);
}

/** callback funkce pro vypis class */
void _cmd_classlist(char *name,objectclass *c) {
 int mu=sizeof(objectclass)+strlen(c->classname)+1+prop_size(c->prop);
 mem_usage+=mu;
 item_count++;
 con_printf("%s [%2d] %s (%d b)",name,c->refcount,modelname(c->model),mu);
}

/** prikaz na vypis nahranych class (debugovaci) */
void cmd_classlist(const char* param) {
 con_printf("Class list:");
 mem_usage=0;item_count=0;
 tree_foreach(class_tree,(tree_func *)_cmd_classlist);
 con_printf("%d bytes, %d items",mem_usage,item_count);
 unused(param);
}

/** callback funkce pro vypis funkci objektu */
void _cmd_ailist(char *name,ai_leaf *c) {
 item_count++;
 con_printf("%08x(%d) %s",c->func,c->type,name);
}

/** prikaz na vypis nahranych funkci objektu(debugovaci) */
void cmd_ailist(const char* param) {
 con_printf("Obj func list:");
 item_count=0;
 tree_foreach(ai_tree,(tree_func *)_cmd_ailist);
 con_printf("%d items",item_count);
 unused(param);
}

//TODO: v tech zvucich by to chtelo obdobu refcount, aby se mohly dostat nekdy i z pameti pryc

//todo: max (soft|hard) (texture|model|sound) memory

//TODO: pokud se hybne kvuli malo FPS o mene nez minframe, tak potom je nutno odecitat rozdily z casu ... aby to pak nebouchlo neco po 2 framech ....

#ifdef have_sound
/** callback funkce pro vypis zvuku */
void _cmd_soundlist(char *name,sound *s) {
 int mu=sizeof(sound)+s->size;
 mem_usage+=mu;
 item_count++;
 con_printf("%s [%d Hz] (%d b)",name,s->freq,mu);
}
#endif

/** prikaz na vypis nahranych zvuku (debugovaci) */
void cmd_soundlist(const char* param) {
 con_printf("Sounds list:");
 mem_usage=0;item_count=0;
 #ifdef have_sound
 tree_foreach(snd_tree,(tree_func *)_cmd_soundlist);
 #endif
 con_printf("%d bytes, %d items",mem_usage,item_count);
 unused(param);
}

/** prikaz na vypis nahranych class (debugovaci) */
void cmd_objlist(const char* param) {
 int i;
 object3d *o;
 vector omodel_abox;
 con_printf("Object list:");
 mem_usage=0;
 for (i=0;i<world->obj.numobj;i++) {
  o=world_getobj(world,i);
  if (o->model) omodel_abox=v_sub(o->model->abox.hi,o->model->abox.lo);
  else {
   v(omodel_abox,-1,-1,-1);
  }
  con_printf("%3d: %g HP %s %s <%g %g %g> (%g %g %g)",i,o->f.health,modelname(o->model),classname(o),all3d(o->p),all3d(omodel_abox));
  if (o->model) mem_usage+=modelsize_mem(o->model);
 }
 con_printf("%d bytes",mem_usage);
 unused(param);
}

/** prikaz na nahrani mapy ve formatu b4d. Bude i s priponou */
void cmd_load(const char* param) {
 worldspawn *w;
 bench_start();
 w=fs_loadworld(param);
 if (w) {
  if (world) {
   world_stop(world);
   freeworld(world);
  }
  world=w;
  world_start(world);
  con_printfv("World %s loaded",param);
 } else {
  con_printf("World %s: load error",param);
 }
 bench_result();
}

/** nahraje uvodni svet - jakysi startup script. Spousti se po nahrani konfigurace a nahrani zakladnich dat */
void b_startup(void){
 con_print(introtext1);//vytisteni infotextu + copyrightu
 con_print(introtext2);
 con_print(introtext3);
 con_print(introtext4);
 //TODO: purge_unused tex/classes/models po zmene levelu, get_id_by_name

 //todo: player starty u gen2d a gen3d
 if (cmd_autostart) {
  cmd_exec(cmd_autostart);
 } else {
  cmd_map("dyna");//TODO: maybe menu main ...
 }
}


/** zobrazi dany svet z pohledu objektu self */
void b_showworld(object3d *self) {
#ifndef dedicated_server
 options opt; 
 glerr("b_showworld:begin");
 polys=0;
 glEnable(GL_BLEND);
 glEnable(GL_CULL_FACE);
 glFrontFace(GL_CW);
// glEnable(GL_TEXTURE_2D);
 opt.head=self->head;
 opt.left=self->left;
 opt.up=self->up;
 opt.pos=self->p;
 gl_initsprites();//vybuchy apod.
 display_worldspawn(world,self,&opt);
 gl_drawsprites(&opt);//vybuchy apod.
 /* if (size2(debug_box.org)>0) {
  glLineWidth(5);
//  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glDisable(GL_CULL_FACE);
  glColor4ub(255,0,0,192);
  glBegin(GL_LINE_LOOP); aa_drawbox(debug_abox); glEnd();
  glColor4ub(255,0,0, 64);
  glBegin(GL_QUADS);     aa_drawbox(debug_abox); glEnd();
  glColor4ub(0,255,0,192);
  glBegin(GL_LINE_LOOP); drawbox(debug_box);  glEnd();
  glColor4ub(0,255,0, 96);
  glBegin(GL_QUADS);     drawbox(debug_box);  glEnd();
 }*/
 glerr("b_showworld:end");
#endif
}

/** na obrazovku vyhodi overlay s FPS informacemi */
void b3d_fps(client *c) {
 dword t;
 if (!fpsdata.num_millis) fpsdata.num_millis=gettickcount();
 fpsdata.num_frames++;
 if (!draw_fps) return;//nic nevykreslovat.
 if ((!c || !c->who) && draw_fps>1) {
  debug_printf0("b3d_fps : No player in one frame\n");
  return;//neni hrac -> stav v menu, mezi levely, atd ....
 }
 t=gettickcount()-fpsdata.num_millis;
 #ifdef vdebug
 if (t<0) fatal("b3d_fps: t<0");
 #endif
 if (t>600) {
  fpsdata.num_fps=(fpsdata.num_frames*1000)/t;
  fpsdata.num_millis=gettickcount();
  fpsdata.num_frames=0;
 }
 if (draw_fps==3) sprintf(fps,"%d fps, x=%5.2f, y=%5.2f, z=%5.2f r=%5.2f,\npolys: %d\n%s",fpsdata.num_fps,c->who->p.x,c->who->p.y,c->who->p.z,c->data.human.r_angle,polys,info);
 else if (draw_fps==2) sprintf(fps,"%d fps, x=%5.2f, y=%5.2f, z=%5.2f\npolys: %d",fpsdata.num_fps,c->who->p.x,c->who->p.y,c->who->p.z,polys);
 else sprintf(fps,"%d fps, %d polys",fpsdata.num_fps,polys);
 if (strlen(fps)==0) return;
#ifndef dedicated_server
 glTranslatex(fontsizex/2.0,fontsizey/2.0,0);
 glColor4ub(0,0,0,128);
 gltext(fps);
 glBlendFunc(GL_ONE,GL_ONE);
 glColor4ub((byte)ceil(255.0*vgamma),0,0,96);
 gltext(fps);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
#else
 printf("%s\n",fps);
#endif
}

#define cannot_shoot(c) ((c)->fire_delay && (c)->fire_time+(c)->fire_delay>gettickcount())//test jestli uz muze vystrelit

/** zobrazi HUD pro jednoho hracre v jeho vyrezu */
void b_oneplayer_hud(client *c) {
 char buf[32];
 int i;
 byte r,g,b;
 int px=c->x1,py=c->y1;
 int wx=c->x2-px,wy=c->y2-py;
 if (wy<1) wy=1;
 b3d_local2d(px,py,wx,wy);
 i=(int)(c->who->f.health+0.99);
 chopmin(i,-999);
 chopmax(i,9999);
 memset(buf,0,sizeof(buf));
 sprintf(buf,ICON_HEALTH "%4d",i);
 hud_show(c,hud_right+hud_bottom,buf,fontsizex,fontsizey,255,255,255,192,0,0,0,0);
 sprintf(buf,"Score: %d " ICON_DEATH " %d",c->frags,c->deaths);
 hud_show(c,hud_left+hud_bottom,buf,fontsizex,fontsizey,255,255,195,192,0,0,0,0);
 if (cannot_shoot(c)) {//aktualni zbran
  sprintf(buf,"(%s)",weapdef[c->weapon].name);r=128,g=128;b=96;
 } else {
  sprintf(buf,"%s",weapdef[c->weapon].name);r=255,g=255;b=192;
 }
 hud_show(c,hud_center+hud_bottom,buf,fontsizex,fontsizey,r,g,b,192,0,0,0,0);
 if (c->who->f.health<=0) {
  hud_show(c,hud_center+hud_vcenter,"KILLED!",-0.1,0,255,0,0,200,100,0,0,150);
 }
 //vytiskne chat
 gl_printchat(c,chatconsole_printstring(),chat_len);
}


/** pro jednoho hrace se zobrazi v jeho okne to, co vidi */
void b_oneplayer_disp(client *c) {
#ifndef dedicated_server
 object3d *p=c->who;
 int px=c->x1,py=c->y1;
 int wx=c->x2-px,wy=c->y2-py;
 if (wy<1) wy=1;
 glerr("b_oneplayer_disp:begin");
 if (!c->who) {
  debug_printf0("b_oneplayer_disp : No player in one frame\n");
  return;//neni hrac -> stav v menu, mezi levely, atd ....
 }
 fpsdata.last_millis=gettickcount();
 b_applymaterial(0);//vyresetuje last-material (napr. barvy se mohly behem menu/hud menit)
 if (use_skybox && current_skymap) b_skybox(px,py,wx,wy,c);// skybox rusi fog
 b3d_frame_single_init(fog_distance);// fog ma kazdy individualne
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 gluPerspective(fov,(bfloat)wx/(bfloat)wy,clipfront,clipback);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 gluLookAt(p->p.x,p->p.y,p->p.z
          ,p->p.x+p->head.x,p->p.y+p->head.y,p->p.z+p->head.z
          ,p->up.x,p->up.y,p->up.z);//nastaveni pohledu
 glViewport(-1+px,-1+py,wx,wy);
 b_showworld(c->who);//vlastni rendering
 b_oneplayer_hud(c);
 glerr("b_oneplayer_disp:end");
#endif
}

/** zkontroluje jestli nedoslo k intermisi a osetri timeout u intermise */
void intermission_check() {
 char buf[1024];
 if (intermis) { //jsme v intermission
  if (intermis->timeout==-1) return;//je vyply rtimeout?
  if (gettickcount()>intermis->timeout+intermis->start_tick) { // pokud doslo k timeoutu
   cmd_exec(intermis->aftercommand); //tak se spusti prikaz
   intermis->timeout=-1;//a vypne timeout, aby se ten prikaz nepoustel furt dokola
  }
 } else { //nejsme v intermission
  int i,n_alive=0,n_all=0,who;
  if (survival) { //survival mode - posledni ziv vyhrava
   //todo: survival s jinym poctem zivotu (int misto bool)
   debug_printf0("Survival mode checking intermission %d\n",screens);
   for (i=0;i<screens;i++) { // projde vsechny obrazovky a hleda hrace
    n_all++;
    if (vplayer[i]->who->f.health>0) { //je zivy?
     n_alive++;who=i;
    }
   }
   if (n_all>1) { // je dosti trapne kdyz jediny hrac hned vyhraje :o)
    if (n_alive==1) { //mame viteze
     _snprintf(buf,sizeof(buf),"Player %d survives with %d frags and wins!",who,vplayer[who]->frags);
     buf[sizeof(buf)-1]=0;//overflow check
     intermission_start("End of game",buf);
    } else if (n_alive==0) { //vsechno pochcipalo
     intermission_start("End of game","Damn! Everybody is dead. Game is tied!");
    }
   }
  }
  if (!fraglimit) return;//fraglimit=0 -> vypnuto
  for (i=0;i<screens;i++) {
   if (vplayer[i]->frags>=fraglimit) { //doshl tenhle fraglimitu?
    _snprintf(buf,sizeof(buf),"Player %d wins with %d frags!",i,vplayer[i]->frags);
    buf[sizeof(buf)-1]=0;//overflow check
    intermission_start("End of game",buf);
    return;
   }
  }
 }
}

/** vola se na konci kazdeho framu - zobrazeni globalniho overlaye prese vsechny okna
 (menu, konzole....) -> po zobrazeni jednotlivych splitscreenu */
void b_postdisplay(void) {
 char ctmp[console_max_cmd+2];
 // tak a ted se nakresli hud/console, atd ....
 glerr("b_postdisplay:begin");
#ifndef dedicated_server
 b3d_2d();
 glPushMatrix();
#endif
 if (screens) b3d_fps(vplayer[0]);//FPS>1 je pouze pro prvniho hrace (globalni)
  else b3d_fps(0);
#ifndef dedicated_server
 glPopMatrix();
 intermission_disp();//zobraz pripadnou intermission screen
 if (game_paused) {//hra je pausnuta, zobraz info
  hud_show(0,hud_center+hud_vcenter," - GAME PAUSED - ",-0.05,0,255,255,255,255,0,0,0,160);
 }
 if (activemenu.type==mtype_simple) {//zobraz menu
  menu_displaysimple(activemenu.menucontent.menu);
 }
 if (activemenu.type==mtype_bind) {
  menu_displaybindeditor();
 }
 if (con->active) {
  glPushMatrix();
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  gl_settexture(tex_console);
  glColor4x(vgamma,vgamma,vgamma,con_alpha);
  glBegin(GL_QUADS);
  glTexCoord2x(0,1-con_height); glVertex3x(0,0,0);
  glTexCoord2x(0,1);		glVertex3x(0,maxy*con_height,0);
  glTexCoord2x(1,1); 		glVertex3x(maxx,maxy*con_height,0);
  glTexCoord2x(1,1-con_height); glVertex3x(maxx,0,0);
  glEnd();
  glTranslatex(fontsizex/4.0,maxy*con_height-fontsizey/4.0-fontsizey,0);
  glColor4x(vgamma*0.75,vgamma*0.75,vgamma*1,con_font_alpha);
  con_draw(console_ofs,(int)((maxy*con_height+fontsizey*0.99)/(bfloat)fontsizey-1));
  glColor4x(vgamma,vgamma,vgamma,con_font_alpha);
  sprintf(ctmp,">%s_",con->cmd);
  gltext(ctmp);
  glPopMatrix();
 }
 glFlush();
 glutSwapBuffers();
#endif
 fpsdata.curr_millis=((gettickcount()-fpsdata.last_millis)+fpsdata.curr_millis)/2;
// if (fpsdata.curr_millis<min_tick) fpsdata.curr_millis=min_tick;
 glerr("b_postdisplay:end");
}

/** vola se na zacatku kazdeho framu - vycisteni z-bufferu a podobne ....
 pred zobrazenim jednotlivych obrazovek */
void b_predisplay(void) {
#ifndef dedicated_server
 b3d_frame_init(fog_color);
#endif
}

int b_handlekey=0;//1 pokud se prave zpracovava nejaka klavesa (tedy jsme v handleru)

/** callback co se vola pri pozadavku na zobrazeni */
cdecl_void b_disp(void) {
 int i;//,j;
 if (b_handlekey) fatalf("b_disp:b_handlekey==1!!!!!!!");
 b_predisplay();
 if (screens) {
  #ifdef have_sound
  b3d_sound_setpos(vplayer[0]->who);//zvuky uslysi pouze prvni hrac
  #endif
  #ifndef dedicated_server
  for (i=0;i<screens;i++) {
   b_oneplayer_disp(vplayer[i]);
  }
  #endif
 } else {
  if (use_skybox && current_skymap) glClear(GL_COLOR_BUFFER_BIT);//je potreba obrazovku vycistit, at to nevypada moc hnusne
  debug_printf0("b_disp : No screens - nothing loaded\n");
  //neni hrac -> stav v menu, mezi levely, atd ....
 }
 b_postdisplay();
}

/** u daneho klienta updatuje rotaci */
void player_updaterotation(client *c) {
 bfloat x,y,z;
 if (c->controller==human) {
  x=c->data.human.r_angle;
  y=c->data.human.r_angle2;
  z=c->data.human.r_angle3;
  c->who->rot.x=-x*180/pi;
  c->who->rot.y=y*180/pi;
  c->who->rot.z=z*180/pi;
 } else fatalf("player_updaterotation");
 //todo: vyresit poradne, a to i pro groundontrol=0
}

/** provede respawn po smrti */
void do_respawn(players *pl) {
 client *c=pl->p;
 debug_printf("do_respawn: begin\n");
 world_kill_object(world,c->who);
 c->who=0;
 addhumanplayer(world,c);//pridame hrace
 if (pl->p->controller==human) {
  pl->moves[use_shoot]=0;//aby hned po respawnu nehazel bomby
 } else debug_printf("do_respawn: NOT A HUMAN!\n");
 debug_printf("do_respawn: end\n");
}

/** zpracuje vystrel daneho klienta */
void do_shot(players *pl,bfloat shot_speed) {
 object3d *shot;
 client *c=pl->p;
 bfloat sb,cb;
 if (cannot_shoot(c)) return;//zbran neumoznuje strilet rychleji
 if (c->who->f.health<=0) {//mrtvi nemohou strilet. Mrtvi se mohou pouze respawnovat
  if (!survival) do_respawn(pl);
  return;
 }
 shot=world_add_object_from(world,c->who);
 if (c->weapon<0 ||c->weapon>=num_weapons) c->weapon=0;
 class_set(shot,weapdef[c->weapon].classname);//classname projektilu
 shot->owner=c->who;
 shot->rotplus.x=5;
 shot->rotplus.z=15;
 shot->rotplus.y=-2;
 sb=sin(bomb_throwup_angle/180*pi);
 cb=cos(bomb_throwup_angle/180*pi);
 v_sum1f(&shot->d,c->who->head,shot_speed*cb);//rychlost strely
 v_addf(&shot->d,c->who->up,shot_speed*sb);//bomb_up
 v_add(&shot->d,c->who->d);//+ rychlost hrace v okamziku vypusteni
 v_addf(&shot->p,c->who->head,8);//zacatek strely od hrace
 c->fire_time=gettickcount();
 c->fire_delay=weapdef[c->weapon].fire_delay;
 world_updateobject(world,shot);
}

/** zpracuje zmenu zbrane daneho klienta */
void do_switch(players *pl,int next,int prev) {
 client *c=pl->p;
 int oldweap=c->weapon;
 if (cannot_shoot(c)) return;//jeste se mu kouri z hlavne
 if (c->who->f.health<=0) return;//mrtvi nemohou strilet
 if (next) c->weapon++; else if (prev) c->weapon--;
 if (c->weapon<0) c->weapon=num_weapons-1;
 if (c->weapon>=num_weapons) c->weapon=0;
 debug_printf("switched %d -> %d\n",oldweap,c->weapon);
 c->fire_time=gettickcount();
 //znemozni strelbu po dobu schovani stare a objeveni nove
 c->fire_delay=weapdef[oldweap].switch_off_delay+weapdef[c->weapon].switch_on_delay;
}

/** odstrani z pameti smazane objekty (co maji nastavenmy flag 'smazany') */
void b_flush_dead_objects(worldspawn *world) {
 int i;
 for(i=world->obj.numobj-1;i>=0;i--) {
  if (world_getobj(world,i)->f.is_destroyed) world_kill_object(world,world_getobj(world,i));//odstraneni
 }
}

/** timer funkce - vola se kazdy frame hry, zajistuje pohyb objektu a vsechno, co je potreba opakovat po kazdy frame */
void b_tick(void){
 object3d *p;
 bfloat r3;
 players *pl=plr;
 bfloat tick_time;

 //spocitani casu mezi dvemi framy
/* now_tick=gettickcount();
 tick_time=(bfloat)(now_tick-last_tick)/1000.0;
 if (tick_time<mintick) tick_time=mintick;//TODO: adjust now_tick/last_tick, pokud bychom se dostavali moc do zaporu, tak sleep();
 if (tick_time>maxtick) tick_time=maxtick;//TODO: issue warning
 last_tick=now_tick;
*/
 tick_time=(bfloat)(fpsdata.curr_millis)/1000.0;

 net_server();//pokud jsem server, zpracuj pakety
 net_client();//pokud jsem klient, zpracuj pakety

 intermission_check();
 if (intermis) return;

 if (!pl || game_paused) return;//nejsou zadni hraci - neni jeste nahrany svet
 while (pl) { //cyklus pro hrace
  p=pl->p->who;
  if (!p) break;//nejsou zadni hraci - neni jeste nahrany svet
  if (flymode) { //cheat
   p->pd.touch=1;
  }
  //CRITICAL TODO: na zaklade bounce/gravitace updatovat touch ....
  p->pd.touch=1;
  if (pl->p->controller==human) {  //pro vsechny human klienty
   if (pl->moves[nextweap] || pl->moves[prevweap]) {//zmena zbrane (ma prednost pred strelbou, kdyz hrac drzi oboji)
    do_switch(pl,pl->moves[nextweap],pl->moves[prevweap]);
   }
   if (pl->moves[use_shoot]) {//vystrelit
    do_shot(pl,bomb_throw_speed);
   }
   p->d=p->nd;
   r3=roll_speed*fpsdata.curr_millis*(pl->moves[pitch_left]-pl->moves[pitch_right]);
   if (groundcontrol) {
    p->head.x=sin(pl->p->data.human.r_angle)*cos(pl->p->data.human.r_angle2);
    p->head.y=cos(pl->p->data.human.r_angle)*cos(pl->p->data.human.r_angle2);
    p->head.z=sin(pl->p->data.human.r_angle2);
    p->left.x=sin(pl->p->data.human.r_angle+pi/2);
    p->left.y=cos(pl->p->data.human.r_angle+pi/2);
    p->left.z=0;
    //TODO: rotate p->left by pl->p->data.human.r_angle3 around p->head
    v_mulv2(&p->up,p->left,p->head);
   } else {
    if (size2(p->left)==0.0 || size2(p->up)==0.0 || size2(p->head)==0.0) {
     p->head.x=1;
     p->left.y=-1;
    }
    v_mulv2(&p->up,p->left,p->head);
    p->head.x+=p->left.x*pl->p->data.human.r_angle;
    p->head.y+=p->left.y*pl->p->data.human.r_angle;
    p->head.z+=p->left.z*pl->p->data.human.r_angle;
    normalize(&p->head);
    v_mulv2(&p->left,p->head,p->up);
    p->head.x+=p->up.x*pl->p->data.human.r_angle2;
    p->head.y+=p->up.y*pl->p->data.human.r_angle2;
    p->head.z+=p->up.z*pl->p->data.human.r_angle2;
    normalize(&p->head);
    v_mulv2(&p->up,p->left,p->head);
    p->up.x+=p->left.x*r3;
    p->up.y+=p->left.y*r3;
    p->up.z+=p->left.z*r3;
    v_mulv2(&p->left,p->head,p->up);
    pl->p->data.human.r_angle=0;
    pl->p->data.human.r_angle2=0;
   }
   b_applyaccel(p,(pl->moves[move_fwd]-pl->moves[move_back]),(pl->moves[move_right]-pl->moves[move_left]),(pl->moves[move_up]-pl->moves[move_down]),fpsdata.curr_millis/1000.0);
   pl->p->data.human.r_angle+=(pl->moves[turn_left]-pl->moves[turn_right])*turn_speed*fpsdata.curr_millis;
   pl->p->data.human.r_angle2+=(pl->moves[turn_down]-pl->moves[turn_up])*turn_speed*fpsdata.curr_millis;
   if (groundcontrol) {
    pl->p->data.human.r_angle3+=r3;
    if (pl->p->data.human.r_angle2>pi/2) pl->p->data.human.r_angle2=pi/2;
    if (pl->p->data.human.r_angle2<-pi/2) pl->p->data.human.r_angle2=-pi/2;
    if (pl->p->data.human.r_angle3>pi/2) pl->p->data.human.r_angle3=pi/2;
    if (pl->p->data.human.r_angle3<-pi/2) pl->p->data.human.r_angle3=-pi/2;
   }
   player_updaterotation(pl->p);
  } else {
   fatalf("b_tick: unknown client type");
  }
  //MAYBE TODO: rotace se nemusi povyst (ale tady se vzdy povede, pac postavy budou krychle)
  pl=pl->next;
 }
 b_run_physics(world,tick_time);
 b_flush_dead_objects(world);
}

/*cdecl_void b_time(int i){
 b_tick();
 glutTimerFunc(tick,b_time,i+1);
}*/

/** idle callback (vola se porad, vzdy po skonceni framu a dozavolani handleru...) */
cdecl_void b_idle(void){
 b_tick();
 glutPostRedisplay();
}

/** ulozi veskere nestaveni (pri ukonceni nebo pri pozadavku) */
void cmd_saveallconfig(const char *param) {
 cmd_saveconfig(default_config_file);
 cmd_savealiases(default_alias_file);
 unused(param);
}

/** Uklizeci rutina. Vola se na konci programu */
void b3d_quit(void){
// glutsetkeyrepeat(GLUT_KEY_REPEAT_ON);
 cmd_saveallconfig(0); 
 fs_close();
 glutDestroyWindow(b3d_win);
 #ifdef have_sound
 b3d_stop_sounds();
 alutExit();
 #endif
 debug_printf("clean shutdown\n");
 exit(EXIT_SUCCESS);
}

/** Z retezce P odskubne prvni (p1) parametr a zbytek (p2). 
 pozor - originalni retezec je modifikovan (je do nej vlozen jeden NULL znak */
void split(char *p,char **p1,char **p2){
 *p1=p;
 *p2=p;
 while ((**p2)&&(!isseparator(**p2))) (*p2)++;
 if (!**p2) return;
 **p2=0;(*p2)++;
 while (isseparator(**p2)) (*p2)++;
}

/** priradi prikaz na klavesu */
void cmd_bind(const char *param) {
 char *parm,*param1,*param2;
 binding *bind;
 int i;
 if (!*param) { //vylistovat bindy
  bindlist(0);
  return;
 }
 parm=strdup(param);
 split(parm,&param1,&param2);
 if (!*param2) {//vylistovat 1 bind
  if (bind0) {
   bind=bind0;
   i=string_to_key(param1);
   debug_printf0("cmd_bind: bind entry %d\n",i);
   while (bind) {
    if (bind->key==i) {
     debug_printf0("cmd_bind: bind found %d\n",i);
     if (bind->action!=-1) {
      debug_printf0("cmd_bind: %s -> %s%d\n",param1,key_cmd[bind->action%k_max],bind->action/k_max);
      con_printf("%s -> %s",param1,b_actionname1(bind->action));
     } else if (bind->command) con_printf("%s -> %s",param1,bind->command);
     else con_printf("%s -> BUG - empty bind",param1);
     return;
    }
    bind=bind->next;
   }
  }
  con_printf("No bind: %s",param1);
  free(parm);
  return;
 }
 //nabindovat prikaz
 bind_key(param1,param2);
 free(parm);
/*
 debug_printf0("cmd_bind: unbinded: %s\n",param1);
 cmd_unbind(param1);
 if (!bind0) {
  bind0=bind9=malloc0(sizeof(binding));
 } else {
  bind9->next=malloc0(sizeof(binding));
  bind9=bind9->next;
 }
 bind9->key=string_to_key(param1);
 debug_printf0("cmd_bind: binding to key: %s -> %d\n",param1,bind9->key);
 //jde to od 1, protoze prvni je NOP
 for (i=1;i<elements(key_cmd);i++) {// je to nejaky jednoduchy command
  if ((j=is_n_cmd(i,param2))>=0) {
  debug_printf0("cmd_bind: binding action: %s -> %d\n",param2,j);
   bind9->action=j;
   free(parm);
   return;
  }
 }
 // je to slozitejsi prikaz;
 bind9->action=-1;
 bind9->command=strdup(param2); 
*/
}

/** smazne alias */
void cmd_unalias(const char *param) {
 char *alias=tree_delv(&alias_tree,param);
 if (alias) {
  free(alias);
  con_printfv("Alias %s deleted",param);
 } else con_printfv("Alias %s does not exist",param);
}

/** smazne alias */
void sys_unalias(const char *param) {
 char *alias=tree_delv(&alias_tree,param);
 if (alias) {
  free(alias);
 }
}

/** priradi alias k prikazu */
void cmd_alias(const char *param) {
 char *parm,*param1,*param2;
 if (!*param) { //vylistovat aliasy
  aliaslist(0);
  return;
 }
 parm=strdup(param);
 split(parm,&param1,&param2);
 if (*param2) {
  sys_unalias(param1);
  tree_add(&alias_tree,param1,strdup(param2));
  con_printfv("Alias %s is set",param1);
  free(parm);
 } else {
  free(parm);
  parm=tree_search(alias_tree,param1);
  if (parm) con_printf("%s = %s",param1,parm);
  else con_printf("Alias %s is not defined",param);
 }
}

/** vola se pri pouziti klavesy s vnitrnim kodem key, down==0 pokud odmacknuta, jinak stisknuta */
int cmd_usebind(int key,int is_down) {
 binding *bind;
 bind=bind0;
 while (bind) {
  if (key==bind->key) {
   if (bind->action>=0) return bind->action;
   if (is_down) cmd_exec(bind->command);
   return k_nothing;
  }
  bind=bind->next;
 }
 return k_nothing;
}

/** zrusi vsechna prirazeni prikazu na klavesy */
void cmd_unbindall(const char *param) {
 while (bind0) {
  bind9=bind0->next;
  free(bind0);
  bind0=bind9;
 }
 bind0=bind9=0;
 unused(param);
}

//TODO: zadny param ale ARGV/ARGC styl .... u cmd_* procedurek

/** prepne na hrace s danym id */
void cmd_playerswitch(const char *param) {
 activeplayer=atoi(param);
 //tODO: better checking!
 activeplayer_check();
}

/** prepe na dalsiho hrace (defaultni input focus) */
void cmd_playernext(const char *param) {
 activeplayer++;
 activeplayer_check();
 unused(param);
}

/** prepe na predchoziho hrace (defaultni input focus) */
void cmd_playerprev(const char *param) {
 activeplayer--;
 activeplayer_check();
 unused(param);
}

/** pausne/unpausne hru */
void cmd_pause(const char *param) {
 game_paused=!game_paused;
 if (game_paused) con_printf("Game paused");
 else con_printf("Game unpaused");
 unused(param);
}

/** zpracovani klavesy pri zmakcnuti ci odmacknuti */
void b_ikey(int c,int is_down) {
 int kf=k_nothing;
 debug_printf0("IN_KEY: %d %d\n",c,is_down);
//konzole je tu natvrdo, aby si to nejaky mamlas nezasekl, kdyz da unbindall a vyskoci z konzole ....
 if (c=='`') kf=k_console;
 //misto use_bind tu byla key_func
 if ((kf==k_console) & is_down) {
  glutIgnoreKeyRepeat(con->active);
  con->active=!con->active;
  return;
 }
 if (con->active){
  if (is_down) editkey(con->cmd,c,console_max_cmd);
 } else if (activemenu.type) {
  if (activemenu.type==mtype_simple) {
   if (is_down) menukey_simple(c);
  } else if (activemenu.type==mtype_bind) {
   if (is_down) menukey_bind(c);
  }
 } else {
  kf=cmd_usebind(c,is_down);
  if (kf) {
   int num=kf/k_max;
   players *pl;
   if (num==0) num=activeplayer;
   pl=player_getn(num);
   if (pl) {
    switch (kf%k_max) {
     case k_pleft:      pl->moves[pitch_left]=is_down;break;
     case k_pright:     pl->moves[pitch_right]=is_down;break;
     case k_left:       pl->moves[turn_left]=is_down;break;
     case k_right:      pl->moves[turn_right]=is_down;break;
     case k_up:         pl->moves[turn_up]=is_down;break;
     case k_down:       pl->moves[turn_down]=is_down;break;
     case k_move_left:  pl->moves[move_left]=is_down;break;
     case k_move_right: pl->moves[move_right]=is_down;break;
     case k_move_up:    pl->moves[move_up]=is_down;break;
     case k_move_down:  pl->moves[move_down]=is_down;break;
     case k_move_fwd:   pl->moves[move_fwd]=is_down;break;
     case k_move_back:  pl->moves[move_back]=is_down;break;
     case k_shoot:      pl->moves[use_shoot]=is_down;break;
     case k_nextweap:   pl->moves[nextweap]=is_down;break;
     case k_prevweap:   pl->moves[prevweap]=is_down;break;
     case k_score:      pl->moves[scoreboard]=is_down;break;
    }
   }
  }
 }
}

/** handler na zmacknuti/odmacknuti tlacitka mysi */
void b_mouse2(int button,int is_down,int x,int y) {
 b_handlekey=1;
 b_ikey(GLUT_MOUSE+button,is_down==GLUT_DOWN);
 b_handlekey=0;
 unused(x);unused(y);
}

/** handler na zmacknuti specialni klavesy */
cdecl_void b_skey(int c,int x,int y) {
 b_handlekey=1;
 b_ikey(-c,1);
 b_handlekey=0;
 unused(x);unused(y);
}

/** handler na odmacknuti specialni klavesy */
cdecl_void b_skeyup(int c,int x,int y) {
 b_handlekey=1;
 b_ikey(-c,0);
 b_handlekey=0;
 unused(x);unused(y);
}

/** handler na zmacknuti klavesy */
cdecl_void b_key(unsigned char c, int x,int y) {
 b_handlekey=1;
 b_ikey(c,1);
 b_handlekey=0;
 unused(x);unused(y);
}

/** handler na odmacknuti klavesy */
cdecl_void b_keyup(unsigned char c, int x,int y) {
 b_handlekey=1;
 b_ikey(c,0);
 b_handlekey=0;
 unused(x);unused(y);
}

/** handler na zmenu stavu okna (minimalizovano/neminimalizovano) */
cdecl_void b_win(int c) {
 window_minimized=c;
 //TODO: do not draw when minimized
};

/** drobne zkrasleni prostredi zavisle na platforma (napr. ikonka ve windows, atd ...) */
void b3d_platform_init(void){
 #ifdef win32
 HANDLE h,m;
 //aby okno neslo zavrit -> kdyz jde, tak to sleti
 h=FindWindow("GLUT",apptitle);
 (void)SetCapture(h);
 (void)SetFocus(h);
 m=GetSystemMenu(h,false);
 (void)EnableMenuItem(m,SC_CLOSE,MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
 #endif
}

/** konec hry */
void cmd_quit(const char *param) {
 b3d_quit();
 unused(param);
}

/** vypise verzi hry na konzoli */
void cmd_version(const char *param) {
 con_print("Version " version);
 unused(param);
}

/** vypise opengl info na konzoli */
void cmd_glinfo(const char *param) {
 con_printf("OpenGl version: %s",glGetString(GL_VERSION));
 con_printf("Gl vendor: %s",glGetString(GL_VENDOR));
 con_printf("Gl renderer: %s",glGetString(GL_RENDERER));
 con_printfz("Gl extensions: %s",glGetString(GL_EXTENSIONS));
 unused(param);
}

/** vycentruje pohled hrace. */
void cmd_center(const char *param) {
//CMD:CENTER 1. parametr je cislo hrace, 0 nebo bez parametru vycentruje vsechny
 object3d *p;
 players *pl=plr;
 int n=atoi(param);
 if (!pl) return;
 while (pl) {
  if (n<=1) {
   p=pl->p->who;
   v(p->head,1,0,0);
   v(p->left,0,1,0);
   v(p->up  ,0,0,1);
   pl->p->data.human.r_angle=pl->p->data.human.r_angle2=0;
  }
  if (n==1) return;
  n--;
  pl=pl->next;
 }
}

/** zaregistruje callbacks (mys, klavesnice, grafika ....) */
void b3d_callbacks (void) {
 glerr("b3d_callbacks:begin");
 //grafika
 glutDisplayFunc(b_disp);
 glutReshapeFunc(b3d_callbackreshape);
 glutIdleFunc(b_idle);
 glutWindowStatusFunc(b_win);
 //timer (TODO: v novem threadu a jinak ... mozna)
 //glutTimerFunc(tick,b_time,0);
 //mys
 glutMotionFunc(b_mouse);
 glutPassiveMotionFunc(b_mouse);
 glutMouseFunc(b_mouse2);
 glutSetCursor(GLUT_CURSOR_NONE);
 b3d_centermouse();
 //klavesnice
 glutKeyboardFunc(b_key);
 glutKeyboardUpFunc(b_keyup);
 glutSpecialFunc(b_skey);
 glutSpecialUpFunc(b_skeyup);
 glutIgnoreKeyRepeat(1);
 glerr("b3d_callbacks:end");
}

/** zmeni ve fullscreenu videorezim (momentalne nefunguje) */
void cmd_vidmode(const char *param) {
 con_printf("TODO: r_vidmode not working (see GLUT docs)");
 return;
 glutGameModeString(param);
 glutEnterGameMode();
}

/** zmeni velikost okna */
void cmd_window_resize(const char *param) {
 glutReshapeWindow(maxx,maxy);
 unused(param);
}

/** funkce volana po zmene velikosti fontu */
void cmd_font_resize(const char *param) {
 b3d_recalc_window();
 unused(param);
}

unsigned int screenshot_serial=0;// seriove cislo screenshotu, postupne se zvysuje, tipajic obrazky s0.tga, s1.tga, atd ....

//TODO: _vsnprintf (nutno nulovat posledni znak retezce), - osetrit

/** tipne screenshot (volitelny parametr=filename) */
void cmd_screenshot(const char *param) {
 char shotname[256];
 tgafile *ss;//screenshot data
 FILE *f;
 int i;
 if (empty(param)) { //vygeneruj jmeno screenshotu automaticky
  for (i=screenshot_serial;i<UINT_MAX;i++) {//najdi prvni v poradi
   _snprintf(shotname,sizeof(shotname),"shot%04d.tga",i);
   if (!fs_exist(shotname)) break;
  }
  screenshot_serial=i;
  if (i>=UINT_MAX) {// prekrocen limit
   for (i=0;i<UINT_MAX;i++) {//zkus cely rozsah
    _snprintf(shotname,sizeof(shotname),"shot%04d.tga",i);
    if (!fs_exist(shotname)) break;
   }
   screenshot_serial=i;
   if (i>=UINT_MAX) {// prekrocen limit - definitivne (stejnak zadny filesystem nezvladne 4 miliardy souboru v jednom adresari ...)
    con_print("Unable to take screenshot! (too many screenshots)");
    return;
   }
  }
 } else {
  b_strncpy(shotname,param,sizeof(shotname));
  //Always append .tga extension //if (!strchr(shotname,'.'))
  b_strncat(shotname,".tga",sizeof(shotname));
 }
 security_check(shotname);
 ss=maybe_malloc1(sizeof(tgaheader)+maxx*maxy*3);
 if (!ss) {
  con_print("Unable to take screenshot! (out of memory)");
  return;
 }
 memset(ss,0,sizeof(tgaheader));//vynuluje hlavicku
 ss->head.type=2;
 ss->head.bpp=24;
 ss->head.x=(short)maxx;
 ss->head.y=(short)maxy;
 glReadPixels(0,0,maxx,maxy,GL_BGR_EXT,GL_UNSIGNED_BYTE,&ss->data);
 f=fopen(shotname,"wb");
 if (!ss) {
  con_printf("Unable to take screenshot! (Can't write to %s)",shotname);
  free(ss);
  return;
 }
 //fs_rgbfix(ss);
 fwrite(ss,1,sizeof(tgaheader)+maxx*maxy*3,f);
 fclose(f);
 free(ss);
 con_printfv("Screenshot taken (%s)",shotname);
}

/** vypise na konzoli pozici a vektory hrace */
void cmd_ppos(const char *param) {
 client *player1=client_getn(atoi(param));
 if (!player1) {
  con_printf("No such player in game: %s",param);
  return;
 }
 con_printf("pos = %g %g %g",all3d(player1->who->p));
 con_printf("head= %g %g %g",all3d(player1->who->head));
 con_printf("up  = %g %g %g",all3d(player1->who->up));
 con_printf("left= %g %g %g",all3d(player1->who->left));
}

/** prida hrace do hry */
void cmd_playeradd(const char *param) {
 addhumanplayer(world,0);
 con_printf("Player added");
 unused(param);
}

/** odstreli jednoho hrace ve hre (s danym cislem, nebo last=posledni)*/
void cmd_playerkill(const char *param) {
 client *c=0;
 int i,j;
 int n;
 if (eqi(param,"last")) n=player_getnum(); else n=atoi(param);
 if (!plr) {
  con_printf("No players exist");
  return;
 }
 if (!plr->next) {
  con_printf("Cannot remove last player");//TODO: osetrit mapu bez hracu (pridat neajky dummy nebo tak nejak)
  return;
 }
 if (n>0) c=client_getn(n);
 if (!c) {
  con_printf("Player %d not found",n);
  return;
 }
 killclient(c);
//TODO: also playershow/playerhide (only change vis)
}

/** nastavi pocet hracu ve hre na danou hodnotu */
void cmd_playersetnum(const char *param) {
 int n,num;
 n=atoi(param);//kolik tam ma byt nove hracu
 if (n<1) {
  con_printf("Must have at least one player");
  return;
 }
 if (n>64) {
  con_printf("64 players is reasonable maximum. Use player_add if you want more");
  return;
 }
 num=player_getnum();
 while (num>n) {//ubirani posledniho hrace dokad jich neni dost
  debug_printf("have: %d, want %d\n",num,n);
  cmd_playerkill("last");
  num--;
 }
 while (num<n) { //pridavani hrace dokad jich neni dost
  debug_printf("have: %d, want %d\n",num,n);
  cmd_playeradd(0);
  num++;
 }
}

#define client_controller(x) (((x)==human)?"human":((x)==camera)?"camera":"?")

/** vypise seznam hracu ve hre */
void cmd_playerlist(const char *param) {
 players *pl=plr;
 int i=1;
 con_printf("Current players:");
 while(pl) {
  con_printf("%2d (%s) frags=%4d, deaths=%4d %s %s",i,client_controller(pl->p->controller),pl->p->frags,pl->p->deaths,pl->p->who->f.health<=0?"(DEAD)":"",pl->p->vis==screen?"visible":"");
  i++;
  pl=pl->next;
 }
 unused(param);
}

/** zaregistruje prikazy a promenne do konzole */
void register_commands(void) {
// glutFullScreen();TODO; dat moznost fullscreen

 register_cmd("generate_3d",&cmd_gen3d,setting,"Generate 3d maze (parameters: X Y Z Corner)");//vygeneruje 3d bludit (parametry: X Y Z Corner)
 register_cmd("generate_2d",&cmd_gen2d,setting,"Generate 2d landscape (parameters: X Y Height)");//vygeneruje 2d krajinu (parametry: X Y Vka)

 //commands
 register_cmd("pause",&cmd_pause,setting,"Pause/unpause the game");//Zastav/znovu pust hru
 register_cmd("r_vidmode",&cmd_vidmode,setting,"Change video mode (BROKEN)");//Zmen video md (NEFUNGUJE)
 register_cmd("centerview",&cmd_center,setting,"Center view");//Vycentruje pohled
 register_cmd("bind",&cmd_bind,setting,"Bind command to keystroke (parameters: Key Command)");//Piad pkaz na klvesu (parametry: Klvesa Pkaz)
 register_cmd("version",&cmd_version,setting,"Print version of program");//Zobraz verzi programu
 register_cmd("glinfo",&cmd_glinfo,setting,"Print info about OpenGL");//Zobraz informace o OpenGL
 register_cmd("unbind",&cmd_unbind,setting,"Delete binding of keystroke to command (parameter: Key)");//Smae piazen pkazu na klvesu (parametr: Klvesa)
 register_cmd("unbindall",&cmd_unbindall,setting,"Delete all keystroke bindings");//Smae vechny piazen klves
 register_cmd("quit",&cmd_quit,setting,"Quit the game");//Ukon hru
 register_cmd("echo",&cmd_echo,setting,"Echo string to console (parameter: string)");//Vype etzec na konzoli (parametr: etzec)
 register_cmd("exec",&cmd_execfile,setting,"Execute a script (parameter: filename)");//Spust skript (parametr: jmno souboru)
 register_cmd("menu",&cmd_menu,setting,"Invoke menu (parameter: name of menu)");//Vyvol menu (parametr: jmno menu)
 register_cmd("clear_history",&cmd_clearhist,setting,"Clear console command history");//Vymae historii pkaz pro konzoli
 register_cmd("clear",&cmd_clear,setting,"Clear console scrollback buffer");//Vymae konzoli
 register_cmd("saveconfig",&cmd_saveconfig,setting,"Save game configuration to specified file (parameter: filename)");//Ulo konfiguraci hry do danho souboru (parametr: jmno souboru)
 register_cmd("savealiases",&cmd_savealiases,setting,"Save aliases to specified file (parameter: filename)");//Ulo aliasy do danho souboru (parametr: jmno souboru)
 register_cmd("saveallconfig",&cmd_saveallconfig,setting,"Save all game configuration to standard files");//Ulo vekerou konfiguraci hry do standardnch soubor
 register_cmd("screenshot",&cmd_screenshot,setting,"Take a screenshot in targa format (optional parameter: filename)");//Sejme obrazovku a ulo ji v targa formtu (nepovinn parametr: jmno souboru)
 register_cmd("map",&cmd_map,setting,"Load new map in .map format (parameter: filename)");//Nahraje novou mapu v .map formtu (parametr: jmno souboru)
 register_cmd("load",&cmd_load,setting,"Load new map in .b4d format (parameter: filename)");//Nahraje novou mapu v .b4d formtu (parametr: jmno souboru)
 register_cmd("tex_list",&cmd_texlist,setting,"List all textures with reference count");//Vype vechny textury vetn potu referenc
 register_cmd("tex_flush",&cmd_flushtex,setting,"Flush all textures with reference count equal to zero (unused textures)");//Uvoln z pamti textury, je maj poet referenc roven nule (nepouit textury)
 register_cmd("list_model",&cmd_modellist,setting,"List all models with reference count");//Vype vechny modely vetn potu referenc
 //TODO: model_flush - unsure jestli to jde
 register_cmd("list_class",&cmd_classlist,setting,"List all classes with reference count");//Vype vechny classes vetn potu referenc
 register_cmd("list_func",&cmd_ailist,setting,"List all object functions");//Vype vechny funkce objekt
 //TODO: class_flush - unsure jestli to jde
 register_cmd("list_obj",&cmd_objlist,cheat,"List all objects in world");//Vype vechny objekty ve he
 //TODO: one object info
 register_cmd("list_sounds",&cmd_soundlist,setting,"List all sounds in memory");//Vype vechny nahran zvuky
 register_cmd("alias",&cmd_alias,setting,"Create a command alias (parameter: Alias Command)");//Vytvo alias k pkazu (parametry: Alias Pkaz)
 register_cmd("unalias",&cmd_unalias,setting,"Deletes a command alias (parameter: Alias)");//Smae alias k pkazu (parametr: Alias)
 register_cmd("help",&cmd_help,setting,"Print helptext for command or variable (parameter: command or variable name)");//Vype strunou npovdu k pkazu nebo k promnn (parametr: jmno promnn nebo pkazu)
 register_cmd("ppos",&cmd_ppos,cheat,"Print player position");//Vype pozici hre ve he
 register_cmd("player_add",&cmd_playeradd,setting,"Add human-controlled player to game");//Pid lovkem zenho hre do hry
 register_cmd("player_kill",&cmd_playerkill,setting,"Remove player with given number from game (parameter: number, last=last player)");//Odstran hre s danm slem ze hry (parametr: slo, last= posledn hr)
 register_cmd("player_setnum",&cmd_playersetnum,setting,"Set number of player. Adds/deletes players as necessary (parameter: number of players)");//Nastav poet hr. Pid, i odebere hre dle poteby (parametr: poet hr)
 register_cmd("player_list",&cmd_playerlist,setting,"List players in game");//Vype seznam hr ve he
 register_cmd("player_next",&cmd_playernext,setting,"Switches to next player");//Pepne se na dalho hre
 register_cmd("player_prev",&cmd_playerprev,setting,"Switches to previous player");//Pepne se na pedchozho hre
 register_cmd("player_use",&cmd_playernext,setting,"Switches to player with given number (parameter: number)");//Pepne na hre s danm slem (parametr: slo)

 register_cmd("menu_add",&cmd_addmenu,setting,"Adds menu with given name and definition (parameters: Name Definition)");//Pid menu s danm jmnem a definic (parametry: Jmno Definice)
 register_cmd("menu_del",&cmd_delmenu,setting,"Deletes menu with given name (parameter: Name)");//Smae menu s danm jmnem (parametr: Jmno)
 register_cmd("menu_list",&cmd_listmenu,setting,"Lists all menus registered in game");//Vype vechny menu registrovan ve he
 register_cmd("menu_bind",&cmd_bindmenu,setting,"Shows menu where you can configure keybinds");//Zobraz menu, kde lze pekonfigurovat klvesy 

 register_cmd("list_net",&cmd_net_connections,setting,"Lists all open and listening network connections");//Vype vechny oteven a naslouchajc sov spojen
 register_cmd("name",&cmd_nickname,setting,"Change or show nickname used in network play (optional parameter: new name)");//Vype, nebo zmn pezdvku pouvanou v sov he
 register_cmd("connect",&cmd_connect,setting,"Connects to a network game (parameter: hostname[:port])");//Pipoj se do sov hry (parametr: hostname[:port])
 register_cmd("listen",&cmd_listen,setting,"Start listening for incoming network connections (optional parameter: port)");//Zane pijmat sov pipojen (voliteln parametr: port)
 register_cmd("disconnect",&cmd_disconnect,setting,"Disconnect from network game (optional parameter: message for server)");//Odpoj se od hry (Voliteln parametr: zprva pro server)
 register_cmd("killserver",&cmd_killserver,setting,"Stop listening for network connections and disconnect all players (optional parameter: message for clients)");//Pestane pijmat sov pipojen a odpoj vechny hre (voliteln parametr: zprva pro hre)
 register_cmd("say",&cmd_say,setting,"Send a chat message to all connected players (parameter: message)");//pole zprvu vem pipojenm hrm (parametr: zprva)

 register_cmd("net_scan",&net_scan,setting,"Scan local network for running servers on given port (or default port if none given). Results are written to console");//Prohled lokln s na bzc servery na danm portu (nebo defaultnm portu, pokud nen dn specifikovn). Vsledky jsou vypsny do konzole.

 //critical commands
 register_cmd("r_skymap",&cmd_change_skymap,critical_setting,"Change skybox (parameter: name of skybox)");//Zmn skybox (parametr: jmno skyboxu)
 //TODO: ve finalni verzi pak zrevidovat co vsechno se uklada do configu a co ne ... cheaty a debug sajrajty by tam jit nemely ...

 //user variables
 register_var("servername",string_var,0,0,&servername,setting,1,"Server name send to clients in network play");//Jmno serveru poslanho hrm pi sov he
 register_var("cmd_autostart",string_var,0,0,&cmd_autostart,setting,1,"Command launched on program startup, after config files are loaded");//Pkaz sputn pi startu programu, po nahrn konfiguranch soubor
 register_var("console_logfile",string_var,0,0,&con_logfile,setting,1,"Log all console messsages to given file (if console_log enabled)");//Zapisuje vechny zprvy na konzoli tak do danho souboru (pokud console_log je zapnuto)
 register_bool("console_log",&con_log,setting,1,"If enabled, all console messages are logged to file (see console_logfile variable)");//Pokud je zapnuto, vechny zprvy na konzoli jsou tak zapisovny do souboru (viz promnn console_logfile)
 #ifdef have_sound
 register_special_var("snd_volume",float_var,0,finf,&snd_volume,cmd_snd_volume,setting,1,"Sound volume");//Hlasitost zvuku
 #endif
 register_special_var("window_x",int_var,40,32767,&maxx,cmd_window_resize,setting,1,"X size of game window in pixels");//X velikost hernho okna v pixelech
 register_special_var("window_y",int_var,30,32767,&maxy,cmd_window_resize,setting,1,"Y size of game window in pixels");//Y velikost hernho okna v pixelech
 register_var("console_history_size",int_var,0,inf,&con_hist_max,setting,1,"Number of console command remembered in history");//Poet zapamatovanch pkaz historie
 register_var("gamma",float_var,0.1,1,&vgamma,setting,1,"Gamma correction (BROKEN)");//Gamma korekce (NEFUNGUJE)
 register_var("console_height",float_var,0.05,1,&con_height,setting,1,"Height of console (relative to screen height)");//Vka konzole (relativn k vce obrazovky)
 register_var("console_alpha",float_var,0,1,&con_alpha,setting,1,"Transparency of console");//Prhlednost konzole
 register_var("console_font_alpha",float_var,0,1,&con_font_alpha,setting,1,"Transparency of console font");//Prhlednost fontu konzole
 register_var("bomb_throwup_angle",float_var,0,90,&bomb_throwup_angle,setting,1,"Angle between direction of the view and direction the bombs are thrown (in degrees) - can be between 0 (throw in direction) and 90 (throw maximally upwards)");//hel mezi smrem pohledu a smrem, kterm se hzej bomby (ve stupnch). Me bt mezi 0 (hzet rovn) a 90 (hzet maximln nahoru)
 register_var("bomb_throw_speed",float_var,0.1,999,&bomb_throw_speed,setting,1,"Speed of the bombs you will throw");//Rychlost hzench bomb
 register_var("r_fps",int_var,0,3,&draw_fps,setting,1,"Draw FPS information");//zobrazovn FPS informac
 register_bool("mouse_unlock",&freemouse,setting,1,"Toggle whether mouse is free to move, or held within window center to allow mouselook");//Umon myi opustit okno, nebo ji dret uprosted okna a tm umoit mouselook.
 register_bool("silent",&silent,setting,0,"Turn verbosity on console on/off");//Zapnut/vypnut extra informac na konzoli
 register_bool("r_skybox",&use_skybox,setting,1,"Turn skybox on/off");//Zapnut/vypnut skyboxu
 register_bool("r_ztrick",&z_trick,setting,1,"Use trick to not clear zbuffer (reduces zbuffer precision)");//Pouit triku na nemazn z-bufferu (sniuje pesnost z-bufferu)
 register_float("sensitivity_h",&h_sensitivity,setting,1,"Horizontal sensitivity of mouse");//Horizontln citlivost myi
 register_float("sensitivity_v",&v_sensitivity,setting,1,"Vertical sensitivity of mouse");//Vertikln citlivost myi
 register_float("sensitivity",&sensitivity,setting,1,"Sensitivity of mouse");//Citlivost myi
 register_special_var("font_x",float_var,6,50,&fontsizex,cmd_font_resize,setting,1,"X Size of console font in pixels");//X velikost konzolovho fontu v pixelech
 register_special_var("font_y",float_var,6,50,&fontsizey,cmd_font_resize,setting,1,"Y Size of console font in pixels");//X velikost konzolovho fontu v pixelech
 register_var("chat_messages",int_var,0,50,&chat_max_len,setting,1,"Maximal number of last chat messages shown in main window");//Maximln poet poslednch zprv zobrazench v hlavnm okn
 register_var("chat_timeout",int_var,0,inf,&chat_timeout,setting,1,"How long are chat messages displayed in main window (miliseconds)");//Jak dlouho jsou zprvy zobrazeny v hlavnm okn (milisekund)

 //game critical variables
 register_var("fov",float_var,5,175,&fov,critical_setting,1,"Field of view");//hel pohledu
 register_var("r_clipfront",float_var,0.0005,finf,&clipfront,cheat,0,"Near clipping distance");//oezvn pohkledu - blzk vzdlenost
 register_var("r_clipback",float_var,0.0005,finf,&clipback,setting,0,"Far clipping distance");//oezvn pohkledu - dalek vzdlenost
 register_var("r_fog_distance",float_var,0,finf,&fog_distance,cheat,0,"Fog distance");//Vzdlenost mlhy
 register_var("r_fog_color",vector_var,0,1,&fog_color,cheat,0,"Fog color");//Barva mlhy
 register_bool("groundcontrol",&groundcontrol,critical_setting,1,"Type of player control (quake-like/6DOF)");//Typ ovldn hre (quake-like/6DOF)
 register_float("speed",&move_speed,critical_setting,0,"Speed of player movement on ground");//Rychlost pohybu hre na zemi
 register_float("speed_air",&move_airspeed,critical_setting,0,"Speed of player movement while in air");//Rychlost pohybu hre ve vzduchu
 //critical server variables
 register_float("gravity",&gravity,critical_setting,1,"Gravity");//Gravitace
 register_var("fraglimit",int_var,0,inf,&fraglimit,critical_setting,1,"Limit of score (frags)");//Limit skre (poet frag)
 register_bool("survival",&survival,critical_setting,1,"If set to 1, nobody can respawn (survival mode)");//Pokud je nastaveno na 1, nikdo se neme respawnovat(survival md)

 //obvious cheats
 register_bool("fly",&flymode,cheat,1,"When set, player can fly");//Pokud je nastaveno na 1, hr me ltat
 register_bool("noclip",&noclip,cheat,1,"When set, no collisions are done");//Pokud je nastaveno na 1, netestuj se kolize
}

/** naalokuje staticka data */
void b3d_static_init() {
 con=malloc0(sizeof(console_data));
 vplayer=malloc1(screens_max*sizeof(client*));
}

/** main funkce pro console aplikace*/
int main(int argc,char** argv){
 #ifdef vdebug
 int i=0;
// system("z /90x60");
 if ((sizeof(vector)-sizeof(vector2))!=i) fatalf("vector: %d<>%d",sizeof(vector),sizeof(vector2));
 if ((sizeof(aa_bbox)-6*sizeof(bfloat))!=i) fatalf("bbox: %d<>6*%d",sizeof(aa_bbox),sizeof(bfloat));
 printf("%s\n",version);
 #endif
 (void)_mkdir(datadir);//pokud adresar neexistuje, pokus o vytvoreni
 //TODO: ukladani do Application_data/$HOME, pokud to jde
 bsrandom();
 b3d_static_init();
 debug_printf0("init()\n");
 b3d_init(640,480,&argc,argv);
 debug_printf0("init() ok\n");
 #ifdef have_sound
 alutInit(&argc,argv);
 debug_printf0("alut init() ok\n");
 b3d_soundinit();
 #endif
 b3d_callbacks();
 b3d_platform_init();
 register_commands();
 fs_init();
 init_menus();
 ai_register_all();
 net_init();
 cmd_execfile("autoexec.cfg");
 silent=0;
 b_loaddata();
 b_startup();
 glutShowWindow();
 debug_printf("main_loop\n");
 glutMainLoop();//this will never return
 return 0;
}

#ifdef WIN32

/** main funkce pro GDI aplikace */
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow) {
 unused(hInstance);
 unused(hPrevInstance);
 unused(nCmdShow);
 unused(lpCmdLine);
 return main (1,&lpCmdLine);//neni to presne, ale tim co mi pristane na prikazove radce se stejne nezabyvam (zatim)
};

#endif
