/*** 
specificke funkce objektu a sveta
*/

#include "b_ai.h"

worldspawn *world;	// Cely svet. Obsahuje vsechny objektu a polygony a vsechno co je prave ve hre [EXT]
//fyzikalni parametry

//TODO: exploze delat lepe
model3d *explosion_model;//jakysi defaultni model na exploze [EXT]

tree *ai_tree;//strom s objekt funkcemi [EXT]

/** nastavi objektu typ a zavola kontruktor classy */
void class_set(object3d *o,const char *name) {
 o->type=b_loadclass(name,1);
 if (o->type) {
  o->f.is_physical=o->type->is_physical;
  //TODO: is_physical prevest na physical_flags (is_touchable | is_blockable | is_blocking)
  o->f.health=o->type->health;
  properties_copy_class_to_object(o->type,o);//zkopiruje properties
 }
 ai_call_create(o);
}

/** zkontroluje jestli lze zavolat objekt funkci (jestli je nastavena) */
int ai_call_check(object3d *obj,ai_func_type func) {
 if (!obj->type) return 0;//bez typu nic nedelaji
 #ifdef vdebug
 if (func<0 || func>=ai_max) fatalf("ai_call: no such function type: %d",func);//tohle se nestava ...
 #endif
 if (!obj->type->ai[func]) return 0;//null funkce, neni co volat
 return 1;
}

/** zaregistruje jednu objekt funkci*/
void ai_register_func(const char *name,aifunc *func,ai_func_type type) {
 ai_leaf *a;
 #ifdef vdebug
 if (type<0 || type>=ai_max) fatalf("ai_register_func: no such function: %d",type);//tohle se nestava ...
 #endif
 a=malloc1(sizeof(ai_leaf)) ;
 a->func=func;
 a->type=type;
 tree_add(&ai_tree,name,a);
}

/** Spocita vzdalenost daneho bodu od daneho bounding boxu*/
vector abox_distance(vector from, aa_bbox *a) {
 vector res;
 if (from.x<a->lo.x) res.x=from.x-a->lo.x; else if (from.x>a->hi.x) res.x=from.x-a->hi.x; else res.x=0;
 if (from.y<a->lo.y) res.y=from.y-a->lo.y; else if (from.y>a->hi.y) res.y=from.y-a->hi.y; else res.y=0;
 if (from.z<a->lo.z) res.z=from.z-a->lo.z; else if (from.z>a->hi.z) res.z=from.z-a->hi.z; else res.z=0;
 return res;
}

/** spocte vzdalenost od objektu, bere v uvahu i bounding box a tak ... (ne jenom stred objektu) a vrati ji jako vektor*/
vector world_distance_v(object3d *o,vector from) {
 vector dist;
 if (o->model) {
  vector dist2;
  dist=abox_distance(from,&o->model->abox);
  dist=v_sub(dist,o->p);//model muze byt posunut
  dist2=v_sub(o->p,from);
  debug_printf0("%g vs %g",(double)v_size(dist),(double)v_size(dist2));
 } else dist=v_sub(o->p,from);
 return dist; //vzdalenost od vybuchu
}

/** spocte vzdalenost od objektu, bere v uvahu i bounding box a tak ... (ne jenom stred objektu) a vrati ji jako cislo */
bfloat world_distance(object3d *o,vector from) {
/* vector dist;
 if (o->model) {
  vector dist2;
  dist=abox_distance(from,&o->model->abox);
  dist=v_sub(dist,o->p);//model muze byt posunut
  dist2=v_sub(o->p,from);
  debug_printf0("%g vs %g",(double)v_size(dist),(double)v_size(dist2));
 } else dist=v_sub(o->p,from);*/
// return v_size(dist); //vzdalenost od vybuchu
 return v_size(world_distance_v(o,from)); //vzdalenost od vybuchu
}


/** zpusobi area damage o hodnote damage, vzdalenosti range, typ damage je dtype a pocatek je v origin.
 source je objekt, kteremu se pripise bod za pripadny kill */
void aic_areadamage(object3d *source,bfloat damage,bfloat range,damage_type dtype,vector origin) {
 int i;
 bfloat len,damnum;
 debug_printf0("BUM from %d!\n",source);
 world_foreach_obj(world,i) {
  if (!world_getobj(world,i)->f.is_destroyed) {//znicene se uz neposkozuji (to by se pak mohlo zacyklit)
   len=world_distance(world_getobj(world,i),origin);
   if (len<range) {// v dosahu vybuchu
    damnum=damage*(1-len/range);//linear damage fallof (0 at damage range, full at 0)
    ai_call_damage(world_getobj(world,i),damnum,dtype,origin,source);//prideli damage
    debug_printf0("damage at %g to %d,%s (amount=%g)!\n",len,world_getobj(world,i),world_getobj(world,i)->type->classname,damnum);
   };
  }
 }
}

/** vytahne z classy hodnotu name (hleda v properties) a vrati ji jako string */
char *aic_getprop(object3d *self,const char *name) {
 return object_getprop(self,name);
}

/** vytahne z classy hodnotu name (hleda v properties) a vrati ji jako bfloat */
bfloat aic_getpropf(object3d *self,const char *name) {
 return object_getpropf(self,name);
}

/** vytahne z classy hodnotu name (hleda v properties) a vrati ji jako bfloat,
 pokud neexistuje, tak vrati defvalue (defaultni hodnotu) */
bfloat aic_getpropfd(object3d *self,const char *name,bfloat defvalue) {
 return object_getpropfd(self,name,defvalue);
}

/** vytahne z classy hodnotu name (hleda v properties) a vrati ji jako int,
 pokud neexistuje, tak vrati defvalue (defaultni hodnotu) */
int aic_getpropid(object3d *self,const char *name,int defvalue) {
 return object_getpropid(self,name,defvalue);
}

/** prida do classy hodnotu key=value */
void aic_addprop(object3d *self,const char *key,const char *value) {
 property_add(&self->prop,key,value);
}

/** zpusobi zanik objektu */
void aic_destroyobject(object3d *who) {
 who->f.is_destroyed=1;
 ai_call_destroy(who);//callback funkce pred zanikem
 who->f.is_destroyed=2;
}

/** zpusobi zanik objektu BEZ volani destruktoru.
 Napriklad pokud se objekt v konstruktoru rozhodne, ze se ani nechce vytvorit
 a jeho destruktor obsahuje nezadouci efekty (gibs, blood, atd ...)*/
void aic_removeobject(object3d *who) {
 who->f.is_destroyed=2;
}

/** pohyb (teleso typu "kus sutru" bez pohonneho systemu - krome gravitace :o) */
void aic_standard_movement(object3d *who,int ticks) {
 bfloat fadefactor;
 if (!ticks) return;
 fadefactor=0.99*exp(ticks*log(0.99)/200.0);//TODO: pocitat lepe
 who->f.copy_d=1;
 who->nd=who->d;

 //take i rotace faduje
 who->rotplus.x*=fadefactor;
 who->rotplus.y*=fadefactor;
 who->rotplus.z*=fadefactor;
 //TODO: tohle fade a pod. by to chtelo dodelat poradne
 v_mul(&who->nd,fadefactor);//TODO: ground fade by melo byt zavisle na tom, jak dlouho trva fade
 v_addf(&who->nd,down,gravity*ticks);//pridava se cast gravitace co za frame vznikla
 who->d=who->nd;
}

//TODO: proc se hrac pohybuje tak blbe?

/** pohyb (teleso typu hrac, neletajici) */
void aic_player_movement(object3d *who,int ticks) {
 //TODO: tu fyziku to bude chtit predelat .. slozite ... dost 
 who->nd=v_mask(who->d,who->pd.movemask,who->nd);
 // v_mul(&who->nd,air_fade);
 v_mul(&who->nd,who->pd.fade*exp(ticks*(log(0.99)/200.0)));//TODO: ground fade by melo byt zavisle na tom, jak dlouho trva fade
// debug_printf("fade=%g * %g\n",who->pd.fade,exp(ticks*(log(0.99)/200.0)));

 // v_addf(&who->d,who->nd,1);
 v_addf(&who->nd,down,gravity*ticks);//pridava se cast gravitace co za frame vznikla
}

/** zmeni tridu z jednoho objektu na jiny.
 Destruktor stareho se nevola, konstruktor noveho ano */
void aic_recast(object3d *who,const char *name) {
 if (who->type) who->type->refcount--;
 class_set(who,name);
}

/** prehraje zvuk, vychazejici z daneho objektu, variability je mezi 0 a 0.5, urcuje promenslivost vysky zvuku */
void aic_playsound(object3d *o,const char *name,bfloat variability) {
#ifdef have_sound
 sound *s=b_loadsound(name);
 b_playsound(s,o,variability);
#endif
 debug_printf("aic_playsound: %s\n",name);
}

/** vytvori na dane pozici novy objekt daneho classname */
object3d *aic_createobject(vector origin,const char *classname) {
 object3d *newo=world_add_object_from(world,0);
 newo->p=origin;
 class_set(newo,classname);
 world_updateobject(world,newo);
 return newo;
}

/** da objektu nahodnou rychlost/rotaci.. udela z nej letici srapnel/gib */
void ai_dogib(object3d *o) {
 o->rot.x=irandom(200);
 o->rot.y=irandom(200);
 o->rot.z=irandom(200);
 o->rotplus.x=(irandom(18)-9);
 o->rotplus.y=(irandom(18)-9);
 o->rotplus.z=(irandom(18)-9); 
 switch (brandom(3)) {
  case 0:o->rotplus.x*=100;
  case 1:o->rotplus.y*=100;
  default:o->rotplus.z*=100;
 }
 o->d.x=irandom(200)-100;
 o->d.y=irandom(200)-100;
 o->d.z=irandom(200)-100;
}

/** objekt (ze) se rozpadne na jednotliv cihliky */
ai_result wall_destroy(AI_DESTROY) {
 object3d *o;
 int i,ngib=(int)(6-self->f.health/40);
 if (ngib<3) ngib=3;
 debug_printf("%d=ngib\n",ngib);
 for (i=0;i<ngib;i++) {
  o=aic_createobject(self->p,"wall_gib");
  ai_dogib(o);
  v_mul(&o->d,ngib);//vice gibu -> rychlejsi
 //TODO: u zdi by mel byt gib dosti solidni a mel by zranovat
 }
 return ai_ok;
 unused(self);
}

/** na timer nijak nereaguje (dummy funkce) */
ai_result timer_nothing(AI_TIMER) {
 debug_printf0("AI:timer_nothing\n");
 return ai_ok;
 unused(self);
 unused(time);
 unused(alive_time);
}

/** na dotyk nereaguje (dummy funkce) */
ai_result touch_nothing(AI_TOUCH) {
 debug_printf0("AI:touch_nothing\n");
 return ai_ok;
 unused(self);
 unused(touch_point);
 unused(touch_slide);
 unused(touch_push);
 unused(target);
}

/** przdn konstruktor */
ai_result create_nothing(AI_CREATE) {
 debug_printf0("AI:create_nothing\n");
 return ai_ok;
 unused(self);
}

/** konstruktor exploze - nastav model na model exploze */
ai_result create_explosion(AI_CREATE) {
 debug_printf0("AI:create_explosion\n");
 self->model=explosion_model;
 return ai_ok;
}

/** Konstruktor zdi. Pokud je nastaveno appearance, je to pravdpodobnost v procentech, e ze bude vytvoena (default je 100) */
ai_result create_wall(AI_CREATE) {
 int appear=aic_getpropid(self,"appearance",100);//PROP::Probability in percent of appearing in the game\Pravdpodobnost objeven se objektu ve he
 if (irandom(100)>=appear) {
  aic_removeobject(self);
  debug_printf0("AI:create_wall: wall disappeared\n");
 }
 return ai_ok;
}

/** Konstruktor mapobjektu - nezniiteln objekt v map s danm modelem. Stejn jako u zdi, i zde je brna v vahu vlastnost appearance */
ai_result create_mapobject(AI_CREATE) {
 char *modelname;
 model3d *m;
 int appear=aic_getpropid(self,"appearance",100);//PROP::Probability in percent of appearing in the game\Pravdpodobnost objeven se objektu ve he
 if (irandom(100)>=appear) {
  aic_removeobject(self);
  debug_printf0("AI:create_mapobject: object disappeared\n");
  return ai_ok;
 }
 modelname=aic_getprop(self,"model");//PROP::Name of model\Jmno modelu
 if (!modelname || !modelname[0]) {
  debug_printf("AI:create_mapobject: no model specified!\n");
  aic_removeobject(self); //neni model -> spacham sebevrazdu
  return ai_ok;
 }
 m=b_loadmodel_r(modelname);
 if (!m) {
  debug_printf("AI:create_mapobject: model %s can't be loaded!\n",modelname);
  aic_removeobject(self); //neni model -> spacham sebevrazdu
  return ai_ok;
 }
 self->f.modelscale=aic_getpropfd(self,"modelscale",1);//PROP::Model scale, real number\Mtko modelu, reln slo
 self->rot.x=aic_getpropfd(self,"angle",0);//PROP::rotation angle (in degrees)\hel rotace (ve stupnch)
 //TODO: pri rotaci by to melo i updatnout bounding box (vetsinou neni treba, ale jsou i specialni pripady)
 if (!modelname || !modelname[0]) {
  debug_printf("AI:create_mapobject: no model specified!\n");
  aic_removeobject(self); //neni model -> spacham sebevrazdu
  return ai_ok;
 }
 if (self->model) fatalf("AI:create_mapobject: object %s already have model!",modelname);
 self->model=m;
 self->abox=parse_abox(aic_getprop(self,"bbox"));//PROP::alternate bounding box of model\Alternativn bounding box modelu
 debug_printf("AI:create_mapobject: success with %s -> %d\n",modelname,m);
 return ai_ok;
}

/** na pokozen nereaguje (tedy nezniiteln objekt) */
ai_result damage_nothing(AI_DAMAGE) {
 debug_printf0("AI:damage_nothing\n");
 return ai_ok;
 unused(self);
 unused(damage);
 unused(damtype);
 unused(offset);
 unused(attacker);
}

/** pi pokozen sn zdrav (health) objektu (standardn chovn), pokud klesne pod nulu, zavol destruktor */
ai_result damage_normal(AI_DAMAGE) {
 debug_printf0("AI:damage_normal\n");
 self->f.health-=damage;
 if (self->client) fatalf("non-player function called on player");
 if (self->f.health<=0) aic_destroyobject(self);
 return ai_ok;
 unused(damtype);
 unused(offset);
 unused(attacker);
}

/** funkce pro pokozen hre, pokud klesne zdrav (health) pod nulu, je hr zabit a tonkovi pipsn bod */
ai_result damage_player(AI_DAMAGE) {
 bfloat health;
 object3d *o;
 debug_printf0("AI:damage_player\n");
 self->f.health-=damage;
 if (self->f.health<=0) {
  if (self->client) {
   self->client->deaths++;
  } else fatalf("player function called on non-player");
  if (attacker && attacker->client) { // nekdo zautocil a je to hrac
   if (attacker==self) self->client->frags--;//sebevrazda
   else attacker->client->frags++;//kill 
   debug_printf0("kill with attacker\n");
  }
  health=self->f.health;
  aic_recast(self,"player_dead");
  self->f.health=health;
  if (self->f.health<-35) {//gib. Mnoho gibu :o)
   int i,ngib=(int)-self->f.health/7;
   if (ngib<3) ngib=3;
   debug_printf("%d=ngib\n",ngib);
   for (i=0;i<ngib;i++) {
    if (brandom(3)) o=aic_createobject(self->p,"gib2");
    else o=aic_createobject(self->p,"gib");
    ai_dogib(o);
    v_mul(&o->d,ngib/3.0);//vice gibu -> rychlejsi
    o->f.modelscale=sqrt(1.0/ngib)*0.6;//cim vice, tim mensi
    //TODO: mozna lehce doladit velikosti a rychlost gibu ....
   }
  } else { //mrtvola
   o=aic_createobject(self->p,"player_corpse");
  }
 }
 return ai_ok;
 unused(damtype);
 unused(offset);
}

/** Pokozen pro objekty, co jsou pi explozi odmrtny dle */
ai_result damage_pushable(AI_DAMAGE) {
 vector x;
 bfloat weight;
 debug_printf("AI:damage_pushable\n");
 if (damtype!=damage_explosion) return ai_ok;//na nevybusny damage to nereaguje
 debug_printf("AI:damage_pushable - an explosion\n");
 weight=aic_getpropfd(self,"weight",10);//PROP::weight of object - the more weight, the less will be object moved by explosions\vha objektu - m t잚, tm mn bude posunovn explozemi
 x=world_distance_v(self,offset);//vektor k mistu vybuchu
 normalize(&x);//normalizace
 v_mul(&x,damage/weight);//aplikace damage a weight
 v_add(&self->nd,x);//pridani k pohybu
 return ai_ok;
 unused(attacker);
}

//TODO: i mrtvoly by mely bit gibovatelne (pri damage>50)

/** Defaultn destruktor - neudl nic (a objekt je pak znien) */
ai_result destroy_nothing(AI_DESTROY) {
 debug_printf0("AI:destroy_nothing\n");
 return ai_ok;
 unused(self);
}

/** Dotyk: objekt se sm zni (pouv se u bomb, kde znien obvykle vyvol explozi) */
ai_result touch_bomb(AI_TOUCH) {
 debug_printf0("AI:touch_bomb\n");
 aic_destroyobject(self);
 return ai_ok;
 unused(touch_point);
 unused(touch_slide);
 unused(touch_push);
 unused(target);
}

/** Pokozen: objekt se sm zni (pouv se u bomb, kde znien obvykle vyvol explozi) */
ai_result damage_bomb(AI_DAMAGE) {
 debug_printf0("AI:damage_bomb %g\n",damage);
 aic_destroyobject(self);
 return ai_ok;
 unused(damage);
 unused(damtype);
 unused(offset);
 unused(attacker);
}

/** zkonvertuje float na string */
char *ftoa(bfloat x) {
 static char ftoa_static_buffer[40];//staticky buffer pro funkci ftoa
 _gcvt(x,30,ftoa_static_buffer);
 return ftoa_static_buffer;
}

//TODO: brat v uvahu i touch funkci
//todo: gib on touch zpomalovani rotace

//todo: postavy by mely byt schopny vybehnout schody

/** Znien bomby zpsob explozi (ve vlastnostech bomby lze nastavit slu a dosah - damage a range) */
ai_result destroy_bomb(AI_DESTROY) {
 object3d *exp;
 bfloat range,damage;
 debug_printf0("AI:destroy_bomb\n");
 damage=aic_getpropfd(self,"damage",150);//PROP::How much damage it will cause\Jak velk pokozen zpsob
 range=aic_getpropfd(self,"range",damage);//PROP::Range of explosion\Dosah exploze
 aic_areadamage(self->owner,damage,range,damage_explosion,self->p);
 exp=aic_createobject(self->p,"explosion");
 aic_addprop(exp,"range",ftoa(range));
 switch(brandom(8)) {
  case 0:aic_playsound(self,"sound/grenade1.wav",0.4);break;
  case 1:aic_playsound(self,"sound/grenade2.wav",0.4);break;
  case 2:aic_playsound(self,"sound/grenade3.wav",0.4);break;
  case 3:aic_playsound(self,"sound/grenade4.wav",0.4);break;
  default:aic_playsound(self,"sound/explo1.wav",0.5);break;
 }
 return ai_ok;
}

/** timer funkce bomby - po danm ase exploduje */
ai_result timer_bomb(AI_TIMER) {
 //TODO: updatovat rotaci podle uhle a doby na zemi ... atd ...
 if (alive_time>=aic_getpropid(self,"bomb_timer",1000)) {//PROP::Bomb timer - delay between throw and explosion\asova bomby - doba mezi hozenm a exploz
  aic_destroyobject(self);//po par vterinach bouchne
 }
 aic_standard_movement(self,time);
 return ai_ok;
}

/** timer funkce exploze - zajist grafick zobrazen vbuchu (pokozen se udluje jinde) */
ai_result timer_explosion(AI_TIMER) {
 int ktime=aic_getpropid(self,"decay_timer",1000);//PROP::Fadeout timer\Fadeout timer
 bfloat alpha=aic_getpropfd(self,"alpha",1);//PROP::Alpha transparency multiplier\Nsobc koeficient transparence
 if (alive_time>=ktime) alive_time=ktime;//aby to tu nepreteklo
 self->f.modelscale=((bfloat)alive_time/(bfloat)ktime)*aic_getpropfd(self,"range",1);;//PROP::Explosion range\Dosah exploze
 self->f.modelalpha=min(255,alpha*(255-(254*alive_time)/ktime));
 if (self->f.modelscale<=0.01) self->f.modelscale=0.01;
 if (alive_time>=ktime) aic_destroyobject(self);//po par vterinach bouchne
 return ai_ok;
 unused(time);
}

/** po danm ase zane objekt plunyle mizet */
ai_result timer_fade(AI_TIMER) {
 int ftime=aic_getpropid(self,"fade_timer",1000);//PROP::Time when object will begin to fade out\as, kdy objekt zane mizet
 int ktime=aic_getpropid(self,"decay_timer",2000);//PROP::Time when object will be removed\as, kdy objekt pln zmiz a bude odstrann
 if (alive_time>=ktime) aic_destroyobject(self);
 else {
  if (alive_time>=ftime) {
   self->f.modelalpha=255-(254*(alive_time-ftime))/(ktime-ftime);
  }
  if (self->f.is_physical) aic_standard_movement(self,time);
 }
 return ai_ok;
}

/** Timer: objekt se nepohybuje vlastn vl, ale je strkn, i pad */
ai_result timer_stone(AI_TIMER) {
 if (self->f.is_physical) aic_standard_movement(self,time);
 return ai_ok;
 unused(alive_time);
}


/** timer: pohyb hre */
ai_result timer_player(AI_TIMER) {
 aic_player_movement(self,time);
 return ai_ok;
 unused(alive_time);
}

/** k dane classe priradi danou funkco pro dany typ */
void ai_assign(objectclass *c,const char *name,ai_func_type type) {
 ai_leaf *l;
 #ifdef vdebug
 if (type<0 || type>=ai_max) fatalf("ai_assign: no such function: %d",type);//tohle se nestava ...
 #endif
 if (!*name)  c->ai[type]=0; //prazdne jmeno -> null funkce
 l=tree_search(ai_tree,name);
 if (!l) {
  con_printf("Function %s does not exist!",name);
  return;
 }
 if (l->type!=type) {
  con_printf("Function %s is of wrong type (is %d, should be %d)",name,l->type,type);
  return;
 }
 c->ai[type]=l->func;
}

#include "b_ai_int.c" //automaticky generovana cast - registrace funkci
