/*** Filesystem functions
 */

#include "b_fs.h"

#define CASESENSITIVITY (2)  //jestli jsou zip archivy case sensitive

char texdir[]="tex";   // podadresar s texturami [EXT]
char mapdir[]="maps";  // podadresar s mapami    [EXT]
 
/** Nahraje soubor ze zip archivu, vrati jeho velikost, pokud *out je null tak se neco nepovedlo */
int fs_getzippedfile(unzFile uf,void **out,const char* filename) {
 char filename_inzip[256];
 unz_file_info file_info;
 if (unzLocateFile(uf,filename,CASESENSITIVITY)!=UNZ_OK) return 0;
 if (unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0)!=UNZ_OK) return 0;
 *out=(void*)maybe_malloc1(file_info.uncompressed_size);
 if (!*out) return 0;
 if (unzOpenCurrentFile(uf)!=UNZ_OK) {
  free(*out);
  return 0;
 }
 if (unzReadCurrentFile(uf,*out,file_info.uncompressed_size)<0) {
  (void)unzCloseCurrentFile(uf);
  free(*out);
  return 0;
 }
 (void)unzCloseCurrentFile(uf);
 return file_info.uncompressed_size;
}

typedef struct _pakfile {//spojak s zip archivy
 unzFile zip;//handle na otevreny zip archiv
// char *name;//jmeno archivu
 struct _pakfile *next;//dalsi
} pakfile;

pakfile *fs_pak;//zip soubory s daty
//TODO: spis by to chtelo na zacatku nacist cele zip soubory a pak chodit pro konkretni soubory do kontretniho archivu

/** vraci true pokud soubor existuje, 0 pokud neexistuje */
int fs_exist(const char *name) {
 struct stat b;
 if (stat(name,&b)) return 0;
 return 1;
}

/**  Otevre zip archiv. 1=uspech, 0=neuspech. */
int fs_open(const char *name) {
 unzFile uf=unzOpen(name);
 pakfile *pk;
 if (uf) {
  pk=malloc1(sizeof(pakfile));
  pk->next=fs_pak;
  pk->zip=uf;
  fs_pak=pk;
  debug_printf("Added zipfile: %s\n",name);
  return 1;
 }
 return 0;
}

/** vrati velikost souboru */
int fsize(const char *f) {
 struct stat b;
 if (stat(f,&b)) return 0;//return 0 if error
 return b.st_size;
}

/** otevre soubor a vrati jeho velikost */
FILE* fopen2(const char *s1,const char *s2,const char *m, uint *filesize){
 FILE *f;
 char str[257];
 char *t;
 b_strncpy(str,s1,sizeof(str));
 b_strncat(str,s2,sizeof(str));
 t=str;
 while (*t) {
  if (*t==fs2) *t=fs;
  t++;
 }
 f=fopen(str,m);
 if (!filesize) return f;
 if (f) *filesize=fsize(str); else *filesize=0;
 return f;
}

char safefilename[257+10];//jmeno zalozniho souboru

//TODO: pozor, fopen2save a fclose2safe nejsou multithread-safe

/** otevre soubor, ale predtim ten stary zazalohuje */
FILE *fopen2safe(const char *s1,const char *s2){
 //TODO: na netware to po sobe ty zalohy nejak nemaze ...
 char str[257];
 char *t;
 safefilename[sizeof(safefilename)-1]=0;
 b_strncpy(str,s1,sizeof(str));
 b_strncat(str,s2,sizeof(str));
 t=str;
 while (*t) {
  if (*t==fs2) *t=fs;
  t++;
 }
 strcpy(safefilename,str);
 strcat(safefilename,".oldXXXXXX");
 mktemp(safefilename);
 unlink(safefilename);
 rename(str,safefilename);
 return fopen2(s1,s2,"wb",0);
}

/** pokud se soubor podari zapsat, tak smazne starou zalohu */
void fclose2safe(FILE *f) {
 if (!f) return;
 debug_printf("fclose2safe(%s)\n",safefilename);
 if (fclose(f)==0) {
  debug_printf("fclose2safe:unlink(%s)\n",safefilename);
  unlink(safefilename);
 }
}

/** nahraje soubor do pameti, vrati velikost */
int loadfile(void **out,const char *name) {
 FILE *f;
 uint s=0;
 security_check(name);
 f=fopen2(datadir "/",name,"rb",&s);
 if (!f) {
  *out=0;
  return 0;
 }
 if (s) {
  *out=maybe_malloc1(s);
  if (*out) {
   s=fread(*out,1,s,f);
  }
 } else (*out)=(void *)0xdeadbeef;//"naalokovany pointer o velikosti 0" -> stejne na nej nejde hrabnout, staci kdyz neni null
 (void)fclose(f);
 debug_printf0("loaded file: %s\n",name);
 return s;
}

/** nahraje soubor, nejdriv z disku, kdyz neni tak z archivu
 vrati velikost v bajtech a pokud neuspeje, vrati 0 a out je nastaven na NULL
*/
int fs_loadfile(void **out,const char *name) {
 int i=loadfile(out,name);
 pakfile *pk;
 if (*out) return i;
 pk=fs_pak;
 while (pk) {
  i=fs_getzippedfile(pk->zip,out,name);
  if (*out) return i;
  pk=pk->next;
 }
 return 0;
}

/** otevre soubor pro zapisovani a naalokuje buffer */
filebuffer *wstream_open(const char *name) {
 filebuffer *f=malloc1(sizeof(filebuffer));
 f->bytes=0;
 f->f=fopen(name,"wb");
 if (!f->f) {
  free(f);
  return 0;
 }
 return f;
}

/** otevre soubor pro zapisovani a naalokuje buffer */
filebuffer *wstream_open2(const char *dir,const char *name) {
 char str[257];
 char *t;
 str[256]=0;
 b_strncpy(str,dir,sizeof(str));
 b_strncat(str,name,sizeof(str));
 t=str;
 while (*t) {
  if (*t==fs2) *t=fs;
  t++;
 }
 return wstream_open(str);
}

/** vyprazdni buffer do souboru */
void wstream_flush(filebuffer *f) {
 if (f->bytes==0) return;
 fwrite(f->buf,f->bytes,1,f->f);
 f->bytes=0;
}

/** ulozi data do bufferu, pri zaplneni buffer protlaci do souboru */
void wstream_save(filebuffer *f,const void *data,int size) {
 int left;
 while (size) {
  if (f->bytes==FILEBUF_LEN) wstream_flush(f);
  left=FILEBUF_LEN-f->bytes;
  if (left>size) left=size;
  memcpy(f->buf+f->bytes,data,left);
  size-=left;
  f->bytes+=left;
 }
 if (f->bytes==FILEBUF_LEN) wstream_flush(f);
}

/** ulozi integer do souboru (buffer) */
void wstream_saveint(filebuffer *f,int i) {
 wstream_save(f,&i,sizeof(int));
}

/** ulozi float do souboru (buffer) */
void wstream_savefloat(filebuffer *f,bfloat ff) {
 wstream_save(f,&ff,sizeof(bfloat));
}

/** ulozi vector do souboru (buffer) */
void wstream_savevector(filebuffer *f,vector v) {
 wstream_save(f,&v,sizeof(vector));
}

/** ulozi retezec (bez ukoncovaci nuly) do souboru (buffer) */
void wstream_savestring(filebuffer *f,const char *s) {
 wstream_save(f,s,strlen(s));
}

/** ulozi delku retezce (integer) a retezec (bez ukoncovaci nuly) do souboru (buffer) */
void wstream_savelstring(filebuffer *f,const char *s) {
 int l;
 l=strlen(s);
 wstream_saveint(f,l);
 wstream_save(f,s,l);
}

/** uzavre soubor, odalokuje data */
void wstream_close(filebuffer *f) {
 wstream_flush(f);
 fclose(f->f);
 free(f);
}

/** otevre ocislovane zip archivy,pocinaje dyna0.zip */
void fs_init() {
 char buf[256];
 int i=0,r;
 do {
  _snprintf(buf,sizeof(buf),"dyna%d.zip",i);
  r=fs_open(buf);
  i++;
 } while(r);
}

/** zavre vsechny otevrene archivy */
void fs_close(void){
 while (fs_pak) {
  (void)unzCloseCurrentFile(fs_pak->zip);
  fs_pak=fs_pak->next;
 }
}

