/***
 OpenGL rendering funkce
 + GL debugovaci funkce
 */
#include "b_gl.h"

bfloat fov=90;          // field of view                                  [EXT]
bfloat clipfront=4;     // near clip plane                                [EXT]
bfloat clipback=3000;   // far clip plane                                 [EXT]
int maxx,maxy;		// rozmery okna na vykreslovani (x=maxx a y=maxy) [EXT]
int polys;		// citac polygonu ve scene                        [EXT]
int b3d_win;		// GLUT handle okna                               [EXT]
bfloat fontsizex=12;    // velikost fontu (nejen) v konzoli - x           [EXT]
bfloat fontsizey=16;	// velikost fontu (nejen) v konzoli - y           [EXT]

texture *tex_font;       //textura pro font                               [EXT]
texture *tex_console;	 //textura pro pozadi konzole                     [EXT]
texture *tex_skybox[6];  //skybox (6 textur)                              [EXT]
ptexture tex_menu[4];	 //textury v menu                                 [EXT]

bfloat coord[6*12]; //souradnice pro skyboxy

vertex2 coord2[6*3]={//souradnice pro skyboxy
{ 1, 0, 0},{ 0, 1, 0},{ 0, 0,-1},  //front
{ 0, 1, 0},{-1, 0, 0},{ 0, 0,-1}, //right
{-1, 0, 0},{ 0,-1, 0},{ 0, 0,-1}, //back
{ 0,-1, 0},{ 1, 0, 0},{ 0, 0,-1},  //left
{ 0, 0, 1},{-1, 0, 0},{ 0, 1, 0},   //up
{ 0, 0,-1},{-1, 0, 0},{ 0,-1, 0},   //down
};

char fps[1280];//buffer na FPS a debugovaci informace                    [EXT]
char info[1024];//buffer na debugovaci informace (r_fps>=2)              [EXT]
type_fpsdata fpsdata;// Data pro FPS kalkulace                           [EXT]

material default_material={ //material - momentalne je barva a textura, ale bude tu casem i vic 
 {255,255,255,255}, //color
 {0.2,0.2,0.2,  1}, //ambient
 {0.8,0.8,0.8,  1}, //diffuse
 {0  ,0  ,0  ,  1}, //specular
 {0  ,0  ,0  ,  1}, //emission
 0,                 //shininess
 ///zbytek je naplnen nulami (nedulezity(
};//defaultni material (defaultni nastaveni)

material *lastmaterial;//posledni pouzity material

/** aplikuje do opengl material. NULL pouze vyresetuje "nacachovany" material. */
void b_applymaterial(material *m) {
 if (m==lastmaterial) return; //pokud tenhle material jiz je pouzit, tak se nic nestane (optimalizace rychlosti)
 lastmaterial=m;
 if (!lastmaterial) return;
 glColor4ubv(m->color);
/* glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,m->ambient);
 glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,m->diffuse);
 glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,m->specular);
 glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,m->emission);
 glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,m->shininess);*/
}

//DEBUGOVACI FUNKCE

/** debugovaci rutina na vypisovani opengl chyb */
#ifdef vdebug //[PASS]
void glerr(char *x) {
 int e=glGetError();//cislo chyby
 char *ee=0;//popis chyby
 if (!e) return;
 if (e==GL_INVALID_ENUM) ee="GL_INVALID_ENUM";
 if (e==GL_INVALID_VALUE) ee="GL_INVALID_VALUE";
 if (e==GL_INVALID_OPERATION) ee="GL_INVALID_OPERATION";
#ifdef GL_TABLE_TOO_LARGE
 if (e==GL_TABLE_TOO_LARGE) ee="GL_TABLE_TOO_LARGE";
#endif
 if (e==GL_OUT_OF_MEMORY) ee="GL_OUT_OF_MEMORY";
 if (e==GL_STACK_OVERFLOW) ee="GL_STACK_OVERFLOW";
 if (e==GL_STACK_UNDERFLOW) ee="GL_STACK_UNDERFLOW";
 if (e==GL_INVALID_ENUM) ee="GL_INVALID_ENUM";
 if (ee) printf("GL_ERROR at %s :%d %s\n",x,e,ee);
  else printf("GL_ERROR at %s :%d\n",x,e);
}
#else //[PASS]
#define glerr //do nothing //[PASS]
#endif //[PASS]

/** nakresli rovnobezny ctyruhelnik s bodem v org a stranami left a down (debugovaci funkce) */
void gluquad(const vector org,const vector left,const vector down) {
 glTexCoord2x(0,0); glVertex3x(org.x              ,org.y              ,org.z              );
 glTexCoord2x(0,1); glVertex3x(org.x+left.x       ,org.y+left.y       ,org.z+left.z       );
 glTexCoord2x(1,1); glVertex3x(org.x+left.x+down.x,org.y+left.y+down.y,org.z+left.z+down.z);
 glTexCoord2x(1,0); glVertex3x(org.x       +down.x,org.y       +down.y,org.z       +down.z);
}

/** Nakresli na obrazovku 3d bounding box. Vicemene debugovaci funkce, jijnak se nepouziva */
void drawbox(const bbox b) {
 vector c2,ndx,ndy,ndz;
 v_sum4(&c2,b.org,b.dx,b.dy,b.dz);
 gluquad(b.org,b.dx,b.dy);
 gluquad(b.org,b.dx,b.dz);
 gluquad(b.org,b.dy,b.dz);
 v_neg(&ndx,b.dx);
 v_neg(&ndy,b.dy);
 v_neg(&ndz,b.dz);
 gluquad(c2,ndx,ndy);
 gluquad(c2,ndx,ndz);
 gluquad(c2,ndy,ndz);
}

/** Nakresli na obrazovku axis-aligned 3d bounding box. Vicemene debugovaci funkce, jijnak se nepouziva */
void aa_drawbox(const aa_bbox b) {
 vector bsiz;
 vector bx,by,bz;
 bsiz=v_sub(b.hi,b.lo);
 v(bx,bsiz.x,0,0);
 v(by,0,bsiz.y,0);
 v(bz,0,0,bsiz.z);
 gluquad(b.lo,bx,by);
 gluquad(b.lo,bx,bz);
 gluquad(b.lo,bz,by);
 v(bx,-bsiz.x,0,0);
 v(by,0,-bsiz.y,0);
 v(bz,0,0,-bsiz.z);
 gluquad(b.hi,bx,by);
 gluquad(b.hi,bx,bz);
 gluquad(b.hi,bz,by);
}

/** vykresli debugovaci axis-aligned bounding box */
void debug_aabox(const aa_bbox a) {
 glLineWidth(5);
 glEnable(GL_BLEND);
 glDisable(GL_CULL_FACE);
 glColor4ub(255,0,0,192);
 glBegin(GL_LINE_LOOP); aa_drawbox(a); glEnd();
 glColor4ub(255,0,0, 64);
 glBegin(GL_QUADS);     aa_drawbox(a); glEnd();
 glColor4ub(0,255,0,192);
}

//RENDEROVACI FUNKCE

 trans_sprite *ts;//seznam s transparentnimi sprity
 int sprites;//pocet spritu v seznamu
 int sprites_max;//maximalni pocet spritu (pokud by mel byt vyssi, pole se musi prealokovat)

/** inicializuje seznam spritu (vyprazdni ho) */
void gl_initsprites(void) {
 if (!sprites_max) sprites_max=16;//neni li nastavena velikost seznamu
 if (!ts) ts=malloc1(sprites_max*sizeof(trans_sprite));//neni li nic naalokovano
 sprites=0;
}

/** porovnavaci funkce pro integrovany quicksort */
cdecl_int _sprite_compare(const trans_sprite *a,const trans_sprite *b) {
 if (a->distance>b->distance) return -1;
 if (a->distance<b->distance) return 1;
 return 0;
}

int last_texture;//posledni pouzita textura [EXT]

/** nastavi texturu (pokud jiz neni nastavena ta sama) */
void gl_texture_on(int id) {
 if (last_texture==id) return; //pokud je stejna, co tam bylo predtim, tak zpet
 if (!last_texture) glEnable(GL_TEXTURE_2D);//pokud byla vyp;nuta, zapnout
 last_texture=id;
 glBindTexture(GL_TEXTURE_2D,id);
}

/** Vypne texturu */
void gl_texture_off() {
 last_texture=0;
 glDisable(GL_TEXTURE_2D);
}

/** nastavi v opengl danou texturu, nebo NULL pokud textura neni */
void gl_settexture(texture *t) {
 if (t) gl_texture_on(t->id); //zapnuti textury
  else gl_texture_off(); //textura neni
}

/** vykresli vsechny sprity v seznamu */
void gl_drawsprites(const options *opt) {
 int i;
 vertex vtx;
 if (!sprites) return;//neni co kreslit;
// debug_printf("sprites=%d\n",sprites);
 qsort(ts,sprites,sizeof(trans_sprite),_sprite_compare);//co bych se matlal se sortem, kdyz je qsort standardni fukce v C.
 glDisable(GL_CULL_FACE);
 b_applymaterial(&default_material);
 for(i=0;i<sprites;i++) {
//  debug_printf("sprite[%d],%g %g %g -> %g\n",i,all3d(ts[i].pos),ts[i].distance);
  glLoadMatrixd(ts[i].matrix);//TODO: tenhle kus kodu je pekne zprasenej, trochu ho to chce proverit ....
  gl_settexture(ts[i].t);
  glColor4ub(ts[i].color[0],ts[i].color[1],ts[i].color[2],ts[i].alpha);
  glBegin(GL_POLYGON);
  v_sum2f(&vtx,opt->left,-1,opt->up,-1);
  v_addf(&vtx,opt->head,ts[i].pos.z);
  glTexCoord2x(0,0);glVertex3x(all3(vtx));
  v_addf(&vtx,opt->left,2);
  glTexCoord2x(1,0);glVertex3x(all3(vtx));
  v_addf(&vtx,opt->up,2);
  glTexCoord2x(1,1);glVertex3x(all3(vtx));
  v_addf(&vtx,opt->left,-2);
  glTexCoord2x(0,1);glVertex3x(all3(vtx));
  glEnd();
 }
 glEnable(GL_CULL_FACE);
}

/** nakresli dany sprite. Respektive vlozi do seznamu k pozdejsimu nakresleni
modelrealpos je pozice stredu spritu.
vx je vektor posunuti spritu oproti pozici modelrealpos (relativni vzhledem k smeru pohledu hrace)
na jedne modelrealpos muze byt poveseno vice spritu (nekolik vrstev exploze, napr.)
 */
void gl_addsprite(const vertex *vx,const options *opt, texture *t,byte color[4],byte alpha,bfloat spritescale,vertex *modelrealpos) {
 vertex realpos/*,vtx*/;
 if (sprites==sprites_max) {//zvetsi kapacitu pole
  sprites_max*=2;
  ts=realloc(ts,sprites_max*sizeof(trans_sprite));
 }
 glGetDoublev(GL_MODELVIEW_MATRIX,ts[sprites].matrix);//vezme aktualni matici
 ts[sprites].pos=*vx;
 ts[sprites].t=t;
 ts[sprites].color[0]=color[0];
 ts[sprites].color[1]=color[1];
 ts[sprites].color[2]=color[2];
 ts[sprites].color[3]=color[3];
 ts[sprites].alpha=alpha;
 realpos=v_sub(*modelrealpos,opt->pos);
 ts[sprites].distance=v_muls(opt->head,realpos)+vx->z*spritescale;
 sprites++;
}

/** zobrazi model. Pri prvni zobrazeni se take mohou delat optimalizace modelu (call list)*/
void displaymodel(model3d *m,object_features *feat,options *opt,vertex *model_position) {
 int i,nvert,j,nv,nf,np=0;
 facecoord *f;
 modelface *mf;
 vertex *vx;
 uchar alpha;//alpha
 if (!m) return;
 glerr("displaymodel:begin");
// exist_or_return(m,"displaymodel");
 f=m->facev;
 mf=m->face;
 nf=m->nface;
 if (m->call_list) {
  glCallList(m->call_list);// jak jednoduche :o)
  polys+=m->call_polys;
 } else {
 #ifdef automatic_call_list
 m->call_list=glGenLists(1);
 glNewList(m->call_list,GL_COMPILE_AND_EXECUTE);
 #endif
 for (i=1;i<=nf;i++) { /// pro kazdy face
  alpha=mf->m->color[3];
  if (feat && feat->modelalpha) {//nastavitelna alpha
   alpha=feat->modelalpha*alpha/255;
  }
  if (alpha) {// uplne pruhledny face se preskoci
   nvert=mf->ncoord;
   if (nvert!=1) {//u spritu se textury delaji pozdeji
    gl_settexture(mf->m->tex);
    b_applymaterial(mf->m);
   #ifdef debug_normal
    #define normal_scale 0.25	//pokud je zaple debug_normal, tak urcije jak velke normaly budou zobrazeny
    glBegin(GL_LINES);
    vx=&m->vert[f->n];
    glVertex3x(vx->x,vx->y,vx->z);
    glVertex3x(vx->x+mf->normal.x*normal_scale,vx->y+mf->normal.y*normal_scale,vx->z+mf->normal.z*normal_scale);
    glEnd();
   #endif
   }
   //TODO: zbavit se asi kompletne ty gammy
   glerr("displaymodel:before glBegin");
   if (nvert==1) {//sprite
    glerr("displaymodel:before sprite");
    np++;
    vx=&m->vert[f->n];
    gl_addsprite(vx,opt,mf->m->tex,mf->m->color,alpha,feat->modelscale,model_position);
    f++;//dalsi bod
    glerr("displaymodel:after sprite");
   } else { //regulerni cara nebo polygon
    if (nvert==3) {
     glBegin(GL_TRIANGLES);
    } else if (nvert==4) {
     glBegin(GL_QUADS);
    } else if (nvert==2) {
     glBegin(GL_LINES);
    } else {
     glBegin(GL_POLYGON);
    }
    np++;
    for(j=0;j<nvert;j++) {
     nv=f->n;
     glTexCoord2x(f->u,f->v);
     vx=&m->vert[nv];
     glVertex3x(vx->x,vx->y,vx->z);
     f++;//dalsi bod
    }
    glEnd();
   }
   glerr("displaymodel:after glEnd");
//   if (!mf->m->tex) glEnable(GL_TEXTURE_2D);
  } else { //preskoceni pruhledneho facul
   f+=mf->ncoord;
  }
  mf++;//dalsi face
 }
 #ifdef automatic_call_list
 glEndList();
 #endif
 m->call_polys=np;
 polys+=np;
 }
/* glerr("displaymodel:bbox begin");
 debug_aabox(m->abox);
 glerr("displaymodel:bbox end");/**/
 glerr("displaymodel:end");
}

#ifdef use_ode_lib		//[PASS]

/** vynasobi matici daty ziskanymi z ODE (rotace + posunuti) */
void gl_transform(vector pos, const dReal *r){
 GLfloat matrix[16];
 matrix[0]=r[0];
 matrix[1]=r[4];
 matrix[2]=r[8];
 matrix[3]=0;
 matrix[4]=r[1];
 matrix[5]=r[5];
 matrix[6]=r[9];
 matrix[7]=0;
 matrix[8]=r[2];
 matrix[9]=r[6];
 matrix[10]=r[10];
 matrix[11]=0;
 matrix[12]=pos.x;
 matrix[13]=pos.y;
 matrix[14]=pos.z;
 matrix[15]=1;
 glMultMatrixf(matrix);
}
#endif				//[PASS]

/** zobrazi jeden objekt, vcetne podrazenych objektu */
void displayobject(object3d *o,object3d *self,options *opt) {
 exist_or_return(o,"displayobject");
 glerr("displayobject:begin");
 glPushMatrix();
 #ifdef use_ode_lib
 if (o->body) { // pouze z dynamickych objektu, staticke nemaji sanci rotovat
  gl_transform(o->p,dBodyGetRotation(o->body));
 } else 
 #endif
 {
  glTranslatex(o->p.x,o->p.y,o->p.z);
  if (o->rot.x) glRotatex(o->rot.x ,0,0,1);
  if (o->rot.z) glRotatex(o->rot.z ,0,1,0);
  if (o->rot.y) glRotatex(o->rot.y ,1,0,0);
 }
 //TODO: doresit poradne rotaci
 if (o->type && o->type->modelscale.x) {//modelscale pro tuto class
  glScalex(all3(o->type->modelscale));
 }
 if (o->f.modelscale!=0 && o->f.modelscale!=1) glScalex(o->f.modelscale,o->f.modelscale,o->f.modelscale);
 if (o->model) displaymodel(o->model,&o->f,opt,&o->p); else
 if (o->type && o->type->model) displaymodel(o->type->model,&o->f,opt,&o->p); //else je to neviditelny (treba trigger a tak ...)
 displayobjectlist(o->child,self,opt);
 glPopMatrix();
 glerr("displayobject:end");
}

/** zobrazi seznam objektu (na kazdy vola displayobject) */
void displayobjectlist(objectlist o,object3d *self,options *opt) {
 object3d **q;
 int i;
 q=o.obj;
 //TODO: proper sort of transparent polys
 for (i=0;i<o.numobj;i++) {
  if (*q!=self) displayobject(*q,self,opt);
  q++;
 }
}

/** zobrazi cely svet (krome self)*/
void display_worldspawn(const worldspawn *w,object3d *self,options *opt) {
 vector v={0,0,0};
 exist_or_return(w,"display_worldspawn");//neni svet k zobrazovani.
 displaymodel(w->m,0,opt,&v);
 displayobjectlist(w->obj,self,opt);
}

//OPENGL extensions funkce

tree *gle_tree;//tree with supported gl extensions           [EXT]

/** rozparsuje seznam OpenGL extensions a nahazi je do BVS stromu */
tree *glext_parse(const char *extlist) {
 tree *t=0;
 const char *x;
 char *y;
 int n;
 if (!extlist) return 0;
 while (*extlist) {
  x=extlist;
  while (*x && *x!=' ') x++;
  n=x-extlist;
  y=malloc1(n+1);
  b_strncpy(y,extlist,n+1);
  tree_add(&t,y,y);
  free(y);
  while (*x && *x==' ') x++;
  extlist=x;
 }
 return t;
}

/** vykresli obdelnik v 2d souradnicich - barva a textura mui byt predem nastaveny */
void gl_rectangle(bfloat x1,bfloat y1, bfloat x2,bfloat y2) {
 glBegin(GL_QUADS);
 glTexCoord2x(0,0);glVertex3x(x1,y2,0);
 glTexCoord2x(0,1);glVertex3x(x1,y1,0);
 glTexCoord2x(1,1);glVertex3x(x2,y1,0);
 glTexCoord2x(1,0);glVertex3x(x2,y2,0);
 glEnd();
}

//VYSTUP TEXTU

/** vyplivne na obrazovku jeden znak a posune matici */
void glchar(int c) {//32x8 znaku ...
 bfloat y=floor(c/32)/8.0;
 bfloat x=((bfloat)(c & 0x1f))/32.0;
 if (c==13) return;
 if (c==10) return;
 glerr("glchar:begin");
 gl_settexture(tex_font); 
 glEnable(GL_BLEND);
 glBegin(GL_QUADS);
 glTexCoord2x(x,1.0-y-1.0/8.0);	  glVertex3x(0.0,fontsizey,0.0);
 glTexCoord2x(x,1.0-y);		  glVertex3x(0.0,0.0,0.0);
 glTexCoord2x(x+1.0/32.0,1-y);      glVertex3x(fontsizex,0.0,0.0);
 glTexCoord2x(x+1.0/32.0,1-y-1.0/8.0);glVertex3x(fontsizex,fontsizey,0.0);
 glEnd();
 glTranslatex(fontsizex,0.0,0.0);
 glerr("glchar:end");
}

/** vyplivne na obrazovky retezec, rozeznama znak newline*/
void gltext(const char *s) {
 size_t t;
//TODO: text je invertovany!!!!!!!! zmenit!!!!
 if (strlen(s)==0) return;
 glerr("text:begin");
 glPushMatrix();
 glPushMatrix();
 for (t=0;t<strlen(s);t++) {
  if (s[t]=='\n') {
   glPopMatrix();
   glTranslatex(0.0,fontsizey,0.0);
   glPushMatrix();
  }
  glchar(s[t]);
 }
 glPopMatrix();
 glPopMatrix();
 glerr("text:end");
}

/** vyplivne text na obrazovce posunuty o zadane souradnice */
void gltext_at(const char *s,bfloat x,bfloat y) {
 glPushMatrix();
 glTranslatex(x,y,0);
 gltext(s);
 glPopMatrix();
}

/** vytiskne text zarovnany na stred obrazovky*/
void gltext_center(const char *s,bfloat y) {
 gltext_at(s,(maxx-strlen(s)*fontsizex)/2,y);
 //TODO: multiline tisknout fakt ve stredu a ne naprosto sejdrem
}

/** nastavi orezavani pri 2d souradnicich na ramecek dany pozici (x,y) a sirkou (wx,wy)
 nastaveni sirky na (0,0) vypne orezavani */
void gl_2dclip(bfloat x,bfloat y,bfloat wx,bfloat wy) {
 int i;
 GLdouble g[4];
 debug_printf0("gl_2dclip:clipping entered\n");
 if (wx==0 && wy==0) { //disable
  for (i=0;i<4;i++) glDisable(GL_CLIP_PLANE1+i);
  return;
 }
 g[0]= 1;g[1]= 0;g[2]=0;g[3]=  -x; glClipPlane(GL_CLIP_PLANE1,g);
 g[0]=-1;g[1]= 0;g[2]=0;g[3]=wx+x; glClipPlane(GL_CLIP_PLANE2,g);
 g[0]= 0;g[1]= 1;g[2]=0;g[3]=  -y; glClipPlane(GL_CLIP_PLANE3,g);
 g[0]= 0;g[1]=-1;g[2]=0;g[3]=wy+y; glClipPlane(GL_CLIP_PLANE4,g);
 for (i=0;i<4;i++) glEnable(GL_CLIP_PLANE1+i);
 debug_printf0("gl_2dclip:clipping leaved\n");
}

/** nastavi 2d projekci o souradnicich (0,0 - maxx,maxy), vypne Z-buffer a backface culling */
void b3d_2d(void) {
 glerr("b3d_2d:begin");
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glDisable(GL_DEPTH_TEST);
 glDisable(GL_CULL_FACE);
 glScalex(2.0/(bfloat)maxx,-2.0/(bfloat)maxy,1.0);
 glTranslatex(-((bfloat)maxx/2.0),-((bfloat)maxy/2.0),0.0);
 glViewport(0,0,maxx,maxy);
 b_applymaterial(&default_material);//reset material
 glerr("b3d_2d:end");
}

/** nastavi 2d projekci o souradnicich (0,0 - wx,wy), vypne Z-buffer a backface culling */
void b3d_local2d(int px,int py,int wx,int wy) {
 glerr("b3d_local2d:begin");
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glDisable(GL_DEPTH_TEST);
 glDisable(GL_CULL_FACE);
 glScalex(2.0/(bfloat)wx,-2.0/(bfloat)wy,1.0);
 glTranslatex(-((bfloat)wx/2.0),-((bfloat)wy/2.0),0.0);
 glViewport(-1+px,-1+py,wx,wy);
 glerr("b3d_local2d:end");
}

#define xmin (1/512.0) //minimalni souradnice textury. Aby byl skybox spojity, melo by to byt asi pul pixelu od okraje
#define xmax (1-xmin)  //maximalni sourdnice textury (viz xmin)

/** vykresli skybox. wx a wy jsou sirka okna kde skybox bude a x a y jeho pozice */
void b_skybox(int x,int y,int wx,int wy,client *c) {
 int i,j;
 object3d *p=c->who;
 glerr("b_skybox:begin");
 for (i=0;i<1/*GL_MAX_LIGHTS*/;i++) glDisable(GL_LIGHT0+i);//povypinat svetla
 glerr("b_skybox:after gl lights");
 for (i=0;i<=5;i++) {//pro kazdy skybox
  for (j=0;j<=2;j++) {//pro kazdou souradnici
   //pro kazdy vertex
   coord[i*12+j+0]=2.0*clipback*(coord2[i*3+0].a[j]+coord2[i*3+1].a[j]+coord2[i*3+2].a[j]);
   coord[i*12+j+3]=2.0*clipback*(coord2[i*3+0].a[j]+coord2[i*3+1].a[j]-coord2[i*3+2].a[j]);
   coord[i*12+j+6]=2.0*clipback*(coord2[i*3+0].a[j]-coord2[i*3+1].a[j]-coord2[i*3+2].a[j]);
   coord[i*12+j+9]=2.0*clipback*(coord2[i*3+0].a[j]-coord2[i*3+1].a[j]+coord2[i*3+2].a[j]);
  }
 }
 glDisable(GL_DEPTH_TEST);
 glDisable(GL_CULL_FACE);
 glDisable(GL_BLEND);
 glDisable(GL_FOG);
 glerr("b_skybox:after gl attributes");
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 if (wy<1) wy=1;
 gluPerspective(fov,(bfloat)wx/(bfloat)wy,0,clipback*6.0);
 glerr("b_skybox:after projection");
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glViewport(x-1,y-1,wx,wy);
 gluLookAt(0,0,0,p->head.x,p->head.y,p->head.z,p->up.x,p->up.y,p->up.z);
 glerr("b_skybox:after modelview");
 b_applymaterial(&default_material);
// glColor3x(vgamma,vgamma,vgamma);
 glerr("b_skybox:before box");
 for (i=0;i<=5;i++) {
  gl_settexture(tex_skybox[i]);
  glBegin(GL_QUADS);
  glTexCoord2x(xmin,xmin);glVertex3x(coord[0+i*12],coord [1+i*12],coord [2+i*12]);
  glTexCoord2x(xmin,xmax);glVertex3x(coord[3+i*12],coord [4+i*12],coord [5+i*12]);
  glTexCoord2x(xmax,xmax);glVertex3x(coord[6+i*12],coord [7+i*12],coord [8+i*12]);
  glTexCoord2x(xmax,xmin);glVertex3x(coord[9+i*12],coord[10+i*12],coord[11+i*12]);
  glEnd();
  glerr("b_skybox:after skypoly");
 } 
 glEnable(GL_DEPTH_TEST);
 glerr("b_skybox:end");
}

/** vytiskne chat window na obrazovku (spodni konec obrazovky)*/
void gl_printchat(client *c,const char *ms,int lines) {
 int px,py,px2,py2;
 px=c->x1;
 py=c->y2-(lines+1)*fontsizey;
 px2=c->x2;
 py2=c->y2-fontsizey;
 glColor4ub(128,128,255,128);
 gl_settexture(0);
 glBegin(GL_POLYGON);
 glVertex3x(px ,py ,0);
 glVertex3x(px2,py ,0);
 glVertex3x(px2,py2,0);
 glVertex3x(px ,py2,0);
 glEnd();
 glColor4ub(255,255,255,192);
 gltext_at(ms,px,py);
}
