#include "config.h"
#ifdef use_ode_lib
#include <ode/ode.h>
#endif
#include "bilbo.h"
#include "b_config.h"
#include "b_types.h"
#include "b_math.h"
#include "b_mem.h"
#include "b_string.h"
#include "b_util.h"

typedef GLbyte  gl4b[4];	//pole 4 bytu
typedef GLbyte  gl3b[3];	//pole 3 bytu
typedef GLfloat gl4fcolor[4];	//pole 4 GLfloatu (barev)
typedef bfloat gl4f[4];		//pole 4 bfloatu
typedef bfloat gl3f[3];		//pole 3 bfloatu

typedef struct _object3d object3d;// skutecna deklarace o kus dal
//funkce objektu pouzite v object3d
typedef int (aifunc)(object3d *self);

typedef struct {//stuktura s rendering-options
 vector head;//head vektor
 vector left;//left vektor
 vector up;//up vektor
 vector pos;//pozice
} options;

typedef struct { //jede transparentni sprite, pouziva se k jejich zarazeni, trideni a zobrazovani. Sprity se zobrazuji po vsech ostatnich objektech
 bfloat distance;    //vzdalenost od 
 vertex pos;         //pozice spritu
 GLdouble matrix[16];//transformacni matice ktera byla aktivni v momentu, kdy se ma sprite zobrazit
 texture *t;         //textura spritu
 byte color[4];      //barva spritu (pouzivaji se jen RGB)
 byte alpha;         //prusvitnost spritu
} trans_sprite;

typedef struct {// Data pro FPS kalkulace
 uint num_fps;//posledni hodnota FPS (nejaktualnejsi)
 uint num_frames;//pocet framu dobrazeny od doby zacatku mereni
 uint num_millis;//doba zacatku mereni (gettickcount)
 uint last_millis;//doba zacatku zobrazovani posledniho framu
 uint curr_millis;//kolik milisekund se zobrazoval dany frame
} type_fpsdata;

//3d rendering

typedef struct { //material - moentalne je barva a textura, ale bude tu casem i vic 
 byte color[4];  //barva textury (r,g,b,alpha)
 GLfloat ambient[4];//GL_AMBIENT
 GLfloat diffuse[4];//GL_DIFFUSE
 GLfloat specular[4];//GL_SPECULAT
 GLfloat emission[4];//GL_EMISSION
 GLfloat shininess;//GL_SHININESS 
 char *name;     //jmeno materialu
 texture *tex;   //pouzivana textura (muze byt null)
 //todo: tady by mely byt i pstatni properties (shininess, ambient, diffuse, multitexturing, atd .... )
 int ref;//pocet referenci na tenhle material. pokud je =0 tak se zlikviduje
} material;

typedef struct _matlist { //spojak materialu
 material *t;             //material
 struct _matlist *next;   //linked list - ukazatel na dalsi. TODO: bylo by lepsi to mit stromem  
} matlist;

#define gl_ext(x) tree_search(gle_tree,x) //test jestli je GL extension podporovana

typedef struct { //per-face data v modelu
 int ncoord;      //pocet vertexu
 vector normal;   //normala facu (pouziva se hlavne na kolize)
 material *m;      //material facu
} modelface;

typedef struct _model model;
typedef struct { //jednoduchy model (bez skeletal animationa podobnych featur)
 int nvert;         //pocet vertexu
 vertex *vert;      //ukazatel na vertexy, nv* vektor
 //TODO: skeletal animace by tu mely ty vertexy trosicku slozitejsi a promenlivejsi
 //      (akorat jeste neni jasny jestli tam neco takovyho budd)
 int nfacev;        //pocet polozek ve faces (cili pocet vertexu ve vsech facech)
 facecoord *facev;  //ukazatel na faces, po skonceni jednoho zacina dalsi, cele se opakuje 
 int nface;         //pocet faces
 modelface *face;   //pocet vertexu, textura a barva pro kazdy face
 int call_list;     //call list (0=nepouzit)
 int call_polys;    //kolik to ma polygonu (kvuli statistice)
 matlist *matl;     //seznam materialu modelu: TODO: odladit, buhvi kde tam ty referecne countery ted litaj!
 aa_bbox abox;      //axis-aligned bounding box modelu
 model *manager;    //odkaz na infromace model manageru o tomhle modelu (reference count, atd..)
} model3d;

typedef struct {        //TMPmodel - pametove neefektivni ale rychly na vytvareni.
 int numvert,maxvert;   //vertexy (x,y,z)
 int numface,maxface;   //faces, v nich per_face
 int numpoint,maxpoint; //per_point data (1 face=3 a vice points)
 facecoord *p;          //u,v mapa a cislo vertexu (numpoint/maxpoint)
 vertex *v;             //(numvert/maxvert)
 modelface *f;          //(numface/maxface)
 matlist *matl;         //seznam materialu modelu
} tmpmodel;

typedef struct _model { //model - pro modelmanager
 int refcount;//kolikrat je pouzit. Pokud nula, lze odstranit z pameti
 model3d *mem; //kde je v pameti
 char *name;//jmeno
};

typedef struct { //seznam objektu s danou kapacitou
 int numobj;    //pocet objektu. Ve hre se meni jen pomerne lehce v jistych mezich.
 int alocobj;   //kapacita alokovaneho pole objektu. Typicky se nebudou prilis menit.
 struct _object3d **obj; //ukazatel na pole ukazatelu na objekty
} objectlist;

typedef struct _objectclass { //trida objektu
 char *classname;             //jmeno tridy
 model3d *model;              //3d reprezentace
 char *definition;            //textova, nerozparsovana definice objektu. Parsuje se az pri prvnim pouziti objektu, pak se zahodi
 aa_bbox *abox;               //axis-aligned bounding boxu vztazeny k poloze
 aifunc *ai[ai_max];          //Vsechny funkce objektu
 struct parent *_objectclass; //odkaz na rodice (nadrazenou tridu objektu)
 int refcount;                //kolik objektu tohoto typu je ve hre
 int is_physical;             //jsou objekty teto tridy fyzicke?
 bfloat health;               //defaultni health objektu
 properties *prop;            //properties
 vector modelscale;           //zmena velikosti modelu vuci defaultni ... 1=puvodni velikost
} objectclass;

/** jeden 3d objekt - jeho vlastnodsti ve hre + ukazatel na 3d model a zavisle objekty */
typedef struct _object3d {
 #ifdef use_ode_lib
 dGeomID geom;   //ode geometry ID (typ geometrie (koule/krychle)+velikost)
 dBodyID body;   //ODE body ID (moment setrvacnosti, hmotnost, rotace, posun, atd ....)
 dMass mass_info;//ODE mass info (moment setrvacnosti, hmotnost ...)
 #endif
 int id;                    //id objektu (pro sitovou hru)
 int lastupdate;            //kdy byl naposledy objekt zmenen
 vector p;                  //pozice objektu (X,Y,Z)
 vector d;                  // vektor pohybu objektu, kterym se objekt chce pohybovat
 vector nd;                 // vektor pohybu objektu konecny (po pridani gravitace, aplikace kolizi ...)
 vector head,left,up;       // vektor pohledu dopredu, doleva a nahoru (jednotkovy, uziti hlavne u hracu)
 // TODO: rotaci vyjasnit a mozna nejak sjednotit s pohybem
 rotation rot,rotplus;      // rotace objektu a jeji zmena (mozna sjednotit s predchozim radkem)
 objectclass *type;         // pointer na tridu objektu
 properties *prop;          // odkaz na hash-table s properties (health, atd ....)
 struct _object3d *parent;  // odkaz na rodicovsky objekt (0 = worldspawn)
 objectlist child;          // podrazene objekty - (napr. k panacovi bude takto pripojena jeho zbran)
 model3d *model;            // 3d reprezentace, pokud je null, bere se type->model (napr. vsechny zbrane jsou stejne, ale func_elevator muze mit kazdy jiny model)
 object_features f;         //specialni vlastnosti objektu. V jedne strukture jsou kvuli load/save
 temp_physics pd;           //docasna data fyzikalniho modelu pro objekt. neukladaji se
 client *client;            //pokud je objekt ovladan hracem, tak je zde struktura hrace, jinak null
 struct _object3d *owner;           // kdo tenhle objekt vytvoril (muze byt null)
 aa_bbox *abox;             // non-null pokud ma jinou axis-aligned bounding box
};

#ifdef use_ode_lib
typedef struct { //pocitani kolizi (ODE)
 dWorldID world;//svet (ODE)
 dSpaceID space;//space (ODE)
 dJointGroupID jgroup;//joint group (ODE)
} collision_struct;
#endif


typedef struct {// cely svet.
 model3d *m;            //staticka architektura
 properties *prop;      // odkaz na hash-table s properties
 objectlist obj;        //pohyblive objekty ve svete
#ifdef use_ode_lib
 collision_struct c;    //struktura urcena hlavne pro pocitani kolizi (ODE)
#endif
} worldspawn;

typedef uint ip_address;// pouze IPv4, s IPv6 to nefunguje (ale to ostatne malo veci)
typedef unsigned short uint_16; //16bitovy unsigned typ
typedef enum { screen, none } client_visibility;//zobrazovani hrace (na obrazovce, nebo nikde), v budoucnu muze byt i zobrazovani do souboru, atd ...
typedef enum { human, bot, camera, network } client_controller;//kdo ovlada hrace. human - clovek na lokalnim pocitaci. bot - umela inteligence na lokalnim pocitaci. camera - kamera na lokalnim pocitaci. network - nejaky objekt na serveru kam jsem pripojen
//typedef struct {		//struktura klienta pro hrace na siti
// net_con_list *c;               //connection klienta (neni to tu potreba)
//} client_network;
typedef struct {		//struktura klienta pro hrace na lokalnim pocitaci (clovek)
 int control;                  // controller ID (
 bfloat r_angle,r_angle2,r_angle3;      // uhly pohledu (x,y,z)
} client_human;
typedef struct  {		//struktura klienta pro hrace na lokalnim pocitaci (bot)
 int ai_level;                 // uroven umele inteligence (skill)
 //... pozdeji zde asi podle poreby pribudou dalsi polozky
} client_bot;
typedef struct {		//struktura klienta pro lokalni kameru
 int style;                 // typ kamery (chase, all field, ...)
 struct _object3d *track;           // jaky objekt sleduje.
 //... pozdeji zde asi podle poreby pribudou dalsi polozky
} client_camera;
typedef struct _client {	//klient - jeden hrac/kamera/robot/sitove pripojeni ...
 client_visibility vis;        // jestli je hrac videt, nebo ne
 client_controller controller; // kdo hrace ovlada
 struct _object3d *who;                // jaky objekt ve hre je hrac (z nej se pak berou pozice kamery a hrac navic sam sebe pri pohledu z vlastnich oci nevidi)
 int x1,y1,x2,y2;              // vyrez obrazovky 
 int split_screen_number;      // cislo obrazovky pri splitsceenu
 int fire_delay;               // doba (v msec) nez bude moct znovu vystrelit
 uint fire_time;                // doba (gettickcount) kdy jsem naposled strilel
 int weapon;                   // aktualne zvolena zbran
 int frags;                    // pocet killu
 int deaths;                   // pocet smrti
 union {                       // bere se zaznam dle polozky controller
//  client_network net;	//sitovy hrac
  client_bot bot;	//bot (umela inteligence)
  client_human human;	//clovek sedici u lokalni klavesnice/mysi
  client_camera cam;	//kamera
 } data;
};

typedef struct _players {// struktura se vsemi LOKALNIMI aktivnimi hraci - spojovy seznem. Obsahuje co kdo ma jak zmacknuto + promenne specificke pro hrace (zbrane a tak ...)
 client *p; //hrac. Jeden hrac nebo kamera ...
 int moves[move_last+1];//struktura s pohybem hrace
 struct _players *next;
} players;
