/*** Inline math funkce. Podle nastaveni kompilatoru se bud soupnou do
 hlavicky, nebo do tela a vygeneruje se hlavicka
*/
/** velikost vektoru umocnena na druhou (rychlejsi nez v_size a pro porovnani staci) */
_inline_ bfloat size2(const vector d) {
 return (d.x*d.x+d.y*d.y+d.z*d.z);
}

/** vektor ze tri cisel */
_inline_ vector vect(bfloat x,bfloat y,bfloat z) {
 vector v;
 v.x=x;
 v.y=y;
 v.z=z;
 return v;
}

/** normalizuje vektor na jednotkovou velikost */
_inline_ void normalize(vector *v) {
 bfloat size=sqrt(size2(*v));
 if (size==0.0) return;
 v->x=v->x/size;
 v->y=v->y/size;
 v->z=v->z/size;
}

/** normalizuje vektor na jednotkovou velikost, prvni nenulova souradnice kladna */
_inline_ void normalize_p(vector *v) {
 bfloat size=sqrt(size2(*v));
 if (size==0.0) return;
 if (v->x<0.0) size=-size; else if (v->x==0.0) {
  if (v->y<0.0) size=-size; else if (v->y==0.0) {
   if (v->z<0.0) size=-size;
  }
 }
 v->x=v->x/size;
 v->y=v->y/size;
 v->z=v->z/size;
}

/** result=-a */
_inline_ void v_neg(vector *result,const vector a) {
 result->x=-a.x;
 result->y=-a.y;
 result->z=-a.z;
}

/** return -a */
_inline_ vector v_negv(const vector a) {
 vector result;
 result.x=-a.x;
 result.y=-a.y;
 result.z=-a.z;
 return result;
}

#define delta 0.0002 //tolerance chyby (relativne vzhledem k "merene velicine")

/** return (a==b) (tolerance chyby) */
_inline_ int v_eq(const vector a,const vector b) {
 return ((fabs(a.x-b.x)<delta) && (fabs(a.y-b.y)<delta) && (fabs(a.z-b.z)<delta));
}

/** return (a==b) vrati nejvetsi rozdil mezi souradnicemi vektoru */
_inline_ bfloat v_eq_delta(const vector a,const vector b) {
 return max(max(fabs(a.x-b.x),fabs(a.y-b.y)),fabs(a.z-b.z));
}

/** return (a==(-b)) (tolerance chyby)*/
_inline_ int v_eq_neg(const vector a,const vector b) {
 return (fabs(a.x+b.x)<delta && fabs(a.y+b.y)<delta && fabs(a.z+b.z)<delta);
}

/** return (a==b) (normalized, positive direction,tolerance chyby ) */
_inline_ int v_eq_np(vector a,vector b) {
 normalize_p(&a);
 normalize_p(&b);
 return (fabs(a.x-b.x)<delta && fabs(a.y-b.y)<delta && fabs(a.z-b.z)<delta);
}

/** return (a==b) (normalized, tolerance chyby ) */
_inline_ int v_eq_n(vector a,vector b) {
 normalize(&a);
 normalize(&b);
 return (fabs(a.x-b.x)<delta && fabs(a.y-b.y)<delta && fabs(a.z-b.z)<delta);
}

/** a=a+b */
_inline_ void v_add(vector *a,const vector b) { 
 a->x+=b.x;
 a->y+=b.y;
 a->z+=b.z;
}

/** a=a+b*c */
_inline_ void v_addf(vector *a,const vector b,bfloat c) { 
 a->x+=b.x*c;
 a->y+=b.y*c;
 a->z+=b.z*c;
}

/** a=a*c */
_inline_ void v_mul(vector *a,bfloat c) {
 a->x*=c;
 a->y*=c;
 a->z*=c;
}

/** return a*b (vector) */
_inline_ vector v_mulv(const vector a,const vector b) {
 vector result;
 result.x=a.y*b.z-a.z*b.y;
 result.y=a.z*b.x-a.x*b.z;
 result.z=a.x*b.y-a.y*b.x;
 return result;
}

/** result:=a*b (vector) */
_inline_ void v_mulv2(vector *result,const vector a,const vector b) {
 result->x=a.y*b.z-a.z*b.y;
 result->y=a.z*b.x-a.x*b.z;
 result->z=a.x*b.y-a.y*b.x;
}

/** return a*b (scalar) */
_inline_ bfloat v_muls(const vector a,const vector b) {
 return a.x*b.x+a.y*b.y+a.z*b.z;
}

/** return a-b */
_inline_ vector v_sub(const vector a,const vector b) {
 vector c;
 c.x=a.x-b.x;
 c.y=a.y-b.y;
 c.z=a.z-b.z;
 return c;
}

/** return a+b */
_inline_ vector v_plus(const vector a,const vector b) {
 vector c;
 c.x=a.x+b.x;
 c.y=a.y+b.y;
 c.z=a.z+b.z;
 return c;
}

/** return a+b*c */
_inline_ vector v_plusf(const vector a,const vector b,bfloat f) {
 vector c;
 c.x=a.x+b.x*f;
 c.y=a.y+b.y*f;
 c.z=a.z+b.z*f;
 return c;
}

/** result=a+b */
_inline_ void v_sum2(vector *result,const vector a,const vector b) {
 result->x=a.x+b.x;
 result->y=a.y+b.y;
 result->z=a.z+b.z;
}

/** result=a+b+c */
_inline_ void v_sum3(vector *result,const vector a,const vector b,const vector c) {
 result->x=a.x+b.x+c.x;
 result->y=a.y+b.y+c.y;
 result->z=a.z+b.z+c.z;
}

/** result=a+b+c+d */
_inline_ void v_sum4(vector *result,const vector a,const vector b,const vector c,const vector d) {
 result->x=a.x+b.x+c.x+d.x;
 result->y=a.y+b.y+c.y+d.y;
 result->z=a.z+b.z+c.z+d.z;
}

//makro je rychlejsi nez funkce;
/** result:=a*a1+b*b1 */
_inline_ void v_sum2f(vector *result,const vector a,bfloat a1,const vector b,bfloat b1) { 
 result->x=a.x*a1+b.x*b1;
 result->y=a.y*a1+b.y*b1;
 result->z=a.z*a1+b.z*b1;
}

/** result:=a*a1 */
_inline_ void v_sum1f(vector *result,const vector a,bfloat a1) { 
 result->x=a.x*a1;
 result->y=a.y*a1;
 result->z=a.z*a1;
}
