/*** sound subsystem */
#include "b_sound.h"

#ifdef have_sound //[PASS]

#pragma comment( lib, "openal32.lib" )
#pragma comment( lib, "alut.lib" )

/** v pripade ze OPENAL vrati chybu, tak ji vytiskne na konzoli */
void alerr(const char *s) {
 int d=alGetError();
 char *q=0;
 if (d) {
  #ifdef AL_INVALID_ENUM
  if (d==AL_INVALID_ENUM) q="AL_INVALID_ENUM";
  #endif
  if (d==AL_INVALID_VALUE) q="AL_INVALID_VALUE";
  debug_printf("OPENAL error %d(%s) at %s\n",d,q?q:"?",s);
 }
}

/** nastavi listenera pomoci objektu */
void b3d_sound_setpos(object3d *listen) {
 ALfloat c[6];//TODO: nejak to zaridit, aby mohlo byt vice listeneru najednou ... nebo tak nejak :o)
 alerr("listener setup");
 c[0]=listen->p.x;
 c[1]=listen->p.y;
 c[2]=listen->p.z;
// debug_printf("b3d_sound_setpos: listener position %g %g %g\n",(double)c[0],(double)c[1],(double)c[2]);
 alListenerfv(AL_POSITION,c);
 alerr("listener position");
 c[0]=listen->d.x;
 c[1]=listen->d.y;
 c[2]=listen->d.z;
// debug_printf("b3d_sound_setpos: listener velocity %g %g %g\n",(double)c[0],(double)c[1],(double)c[2]);
 alListenerfv(AL_VELOCITY,c);
 alerr("listener velocity");
 c[0]=listen->head.x;
 c[1]=listen->head.y;
 c[2]=listen->head.z;
 c[3]=listen->up.x;
 c[4]=listen->up.y;
 c[5]=listen->up.z;
// debug_printf("b3d_sound_setpos: listener orientation %g %g %g - %g %g %g\n",(double)c[0],(double)c[1],(double)c[2],(double)c[3],(double)c[4],(double)c[5]);
 alListenerfv(AL_ORIENTATION,c);
 alerr("listener orientation");
}

/** nahraje zvuk do pameti a do openal */
sound *fs_loadsound(const char *name) {
 sound *sn;
 void *data;
 if (!fs_loadfile(&data,name)) return 0;
 sn=malloc1(sizeof(sound));
 alGetError();
 alGenBuffers(1,&sn->buffer);
 if (alGetError()!=AL_NO_ERROR) {
  free(sn);
  return 0;
 }
 sn->mem=0;
 alutLoadWAVMemory(data,&sn->format,&sn->mem,&sn->size,&sn->freq,&sn->loop);
 if (sn->mem) {//nejaky data se nahraly
  alBufferData(sn->buffer,sn->format,sn->mem,sn->size,sn->freq);
  alutUnloadWAV(sn->format,data,sn->size,sn->freq);
  alerr("fs_loadsound");
  return sn;
 } else free(sn);
 return 0;
}

tree *snd_tree;//strom se zvuky [EXT]

/** nahraje zvuk do pameti pokud tam neni, pokud tam je tak vrati pointer */ 
sound *b_loadsound(const char *name) {
 sound *s=tree_search(snd_tree,name);
 if (s) return s;
 s=fs_loadsound(name);
 tree_add(&snd_tree,name,s);
 return s;
}

ALuint *snd;//prave prehravane zvuky
int snd_max_sources=0;//kapacita bufferu
int snd_num_sources=0;//pocet zvuku v bufferu

/** jako alGetSourcei, ale funkce hodnotu vraci primo, ne odkazem */
int alGetSource(ALuint src,ALenum name) {
 int r=0;
 alGetSourcei(src,name,&r);
 if (r!=AL_PLAYING) debug_printf("AL state=%d\n",r);
 return r;
}


/** prehraje zvuk z daneho zdroje. Variability: promenlivost vysku zvuku (0-0.5) */
void b_playsound(sound *s,object3d *source,bfloat variability) {
 int i;
// ALuint dummy;
 alerr("b_playsound:begin");
 if (!snd_max_sources) {
  snd_max_sources=4;
  snd=malloc1(sizeof(ALuint)*snd_max_sources);
  alGenSources(snd_max_sources,snd);
 }
 if (snd_num_sources<snd_max_sources) {
  i=snd_num_sources;
  snd_num_sources++;
 } else {//no free buffers
  for(i=0;i<=snd_max_sources;i++) {
   if (i==snd_max_sources) break;
   if (alGetSource(snd[i],AL_SOURCE_STATE)!=AL_PLAYING) break;//buffer not active, reuse
  }
  if (i==snd_max_sources) {//no available buffers
   snd_max_sources*=2;
   snd=realloc1(snd,sizeof(ALuint)*snd_max_sources);
   alGenSources(snd_max_sources/2,snd+snd_max_sources/2);//generate rest of buffers
   snd_num_sources++;
  }
 }
 debug_printf("b_playsound: sound source position %g %g %g\n",all3d(source->p));
 alerr("b_playsound:beginplay");
 if (variability) {
  debug_printf("b_playsound:variability=%g\n",variability);
  variability=1.0+variability*((irandom(100)-50)/100.0);
  debug_printf("b_playsound:pitch=%g\n",variability);
  alSourcef(snd[i],AL_PITCH,variability);
 } else alSourcef(snd[i],AL_PITCH,1.0);
 alerr("b_playsound:pitch");
 alSourcef(snd[i],AL_GAIN,1.0);
 alerr("b_playsound:gain");
 alSource3f(snd[i],AL_POSITION,source->p.x,source->p.y,source->p.z);
 alerr("b_playsound:position");
 alSource3f(snd[i],AL_VELOCITY,0,0,0);//all3(source->d));
 alerr("b_playsound:velocity");
// alSource3f(snd[i],AL_DIRECTION,all3(source->head));
// alerr("b_playsound:direction");
//AL_REFERENCE_DISTANCE to nezna - ???
// alSourcei(snd[i],AL_REFERENCE_DISTANCE,300);
// alerr("b_playsound:reference_distance");
 alSourcei(snd[i],AL_BUFFER,s->buffer);
 alerr("b_playsound:buffer");
 alSourcei(snd[i],AL_LOOPING,AL_FALSE);
 alerr("b_playsound:looping");
 alSourcePlay(snd[i]);
 alerr("b_playsound:sourceplay");
}

bfloat snd_volume=1;//hlasitost zvuku [EXT]
/** nastavi hlasitost zvuku */
void cmd_snd_volume(const char *param) {
 alListenerf(AL_GAIN,snd_volume);
 unused(param);
}

//TODO: ty zvuky to nechce nejak hrat stereo, dokonce i vzdalrenost to ignoruje!!!!

/** inicializace zvukoveho systemu */
void b3d_soundinit() {
alDopplerVelocity(64/2*343);
// alDopplerVelocity(64/2*343);//TODO: pak zmenit jednotku aby vse bylo per second
}

/** zastavi vsechny zvuky */
void b3d_stop_sounds() {
 alerr("b3d_stop_sounds:begin");
 if (!snd_num_sources) return;//no sounds playing
 alSourceStopv(snd_num_sources,snd);
 alerr("b3d_stop_sounds:end");
}

/** deinicializace zvukoveho systemu */
void b3d_soundexit() {
 alDeleteSources(snd_max_sources,snd);
//TODO: alDeleteBuffers((all sound buffers)NUM_BUFFERS, g_Buffers);
}

#endif //[PASS]
