/*** prace s retezci (char *) */
#include "b_string.h"
#include "b_mem.h"
#define STR_HEAD 1 //o kolik je retezec delsi v pameti nez kolik ma znaku

/** vrati 1 pokud je retezec NULL nebo prazdny */
int empty(const char *a) {
 if (!a) return 1;
 return *a==0;
}

/** vrati 1 pokud se retezce rovnaji */
int eq(const char *a,const char *b) {
 if (empty(a)) return (empty(b));
 if (empty(b)) return 0;
 return (strcmp(a,b)==0);
}

/** vrati 1 pokud se retezce rovnaji (case insensitive) */
int eqi(const char *a,const char *b) {
 if (empty(a)) return (empty(b));
 if (empty(b)) return 0;
 return (stricmp(a,b)==0);
}

/** alokace noveho retezce v pameti. Naalokuje retexec pro chars znaku*/
char *str_new(size_t chars) {
 char *x=malloc1(chars+STR_HEAD);
 if (x) *x=0;
 return x;
}

/** realokace retezce na tolik, kolik v pameti skutecne zabira */
void str_exact(char **s) {
 *s=realloc1(*s,strlen(*s)+STR_HEAD);
}

/** odalokace retezce s kontrolou na NULL pointery */
void str_free(char *s) {
 if (s==NULL) return;
 free(s);
}

/** duplikace retezce */
char *strdup(const char *s){
 size_t i=strlen(s);
 char *t=malloc1(i+STR_HEAD);
 strcpy(t,s);
 t[i]=0;
 return t;
}

/** jako strncpy, ale posledni znak vzdy nastavi na NULL */
char *b_strncpy(char *dest,const char *source, size_t count) {
 char *q=strncpy(dest,source,count);
 q[count-1]=0;
 return q;
}

/** jako strncat, ale pocet znaku je omezeni delky retezce celkem, vcetne NULL znaku, ne jen pridavaneho retezce */
char *b_strncat(char *dest,const char *source, size_t count) {
 return strncat(dest,source,count-strlen(dest)-1);
}

/** nahradi v retezci jeden znak jinym */
void str_replace(char *s,char c,char r) {
 while (*s) {
  if (*s==c) *s=r;
  s++;
 }
}

/** priplacne k retezci s znak c, pokud by tim neprekrocil delku maxlen */
void str_addc(char *s,char c,size_t maxlen) {
 size_t l=strlen(s);
 if (l>=maxlen-1) return;
 s[l]=c;
 s[l+1]=0;
}

/** vraci pocet znaku "target" v retezci "string" */
int str_count(const char *string,char target) {
 int i=0;
 if (!string) return 0;
 while (*string) {
  if (*string==target) i++;
  string++;
 }
 return i;
}

/** vezme prvnich n znaku z retezce s a vrati je v novem naalokovanem retezci*/
char *strndup(const char *s,int n) {
 char *x=malloc1(n+1);
 strncpy(x,s,n);
 x[n]=0;
 return x;
}

/** z retezce vecme cast az po prvni delimiter (libovolny znak z retezce delim)
 vrati nove alokovany retezec.*/
char *str_chop_until(char **def,const char *delim) {
 char *p=*def,*q;
 while (!strchr(delim,*p)) {//chci najit prvni delimiter
  if (!*p) break;//end of string
  p++;
 }
 q=strndup(*def,p-*def);//alokovany retezec co chci vratit
 if (*p) p++;//retezec ukazuje bud na konec (0) nebo ho posuneme na znak za delimiter
 *def=p;//oriznu retezec 
 return q;
}

/** prevedeni cisla na retezec (v desitkove soustave) */
char *str_itoa(int n,char *buf) {
#ifdef _WIN32	// verze s itoa
 return itoa(n,buf,10);
#else
 sprintf(buf,"%d",n);//proste zavolame sprintf, na platformach, ktere itoa neznaji
 return buf;
#endif
}

/** znak # v ntemplate nahradi cislem j a vrati novy retezec */
char *str_nreplace(const char *ntemplate,int n) {
 char *res; 
 int m=n,len=strlen(ntemplate),digit=1;
 int p=((int)strchr(ntemplate,'#')-(int)ntemplate);
 while (m>10) {
  m/=10;
  digit++;
 }
 res=malloc1(len+digit/*-1+1*/);
// itoa(n,res+p,10);
 str_itoa(n,res+p);
 strncpy(res,ntemplate,p);
 strncpy(res+p+digit,ntemplate+p+digit,len-p);
 res[len+digit-1]=0;
 return res;
}

/** vraci true pokud retezec str konci retezcem what (case insensitive)*/
int str_endwithi(const char *str,const char *what) {
 int l1=strlen(str);
 int l2=strlen(what);
 if (l1<l2) return 0;//what je vetsi ...
 return eqi(str+l1-l2,what);
}

/** vrati nove naalokovany retezec, obsahujici druhy retezec prilepeny za prvnim (podobne jako strcat)*/
char *str_plus (const char *str1,const char *str2) {
 char *res;
 int len1,len2;
 if (empty(str2)) return strdup(str1);
 if (empty(str1)) return strdup(str2);
 len1=strlen(str1);
 len2=strlen(str2);
 res=str_new(len1+len2);
 strcpy(res,str1);
 strcpy(res+len1,str2);
 res[len1+len2]=0;
 return res;
}

