/***

b_types

Zde jsou definovany typy

*/
#ifndef __B_TYPES_H__ //include only once
#define __B_TYPES_H__

// type aliases & simple types

typedef unsigned int dword;  //32 bitovy unsigned typ
typedef unsigned int uint;   //32 bitovy unsigned typ
typedef unsigned char byte;  // 8 bitovy unsigned typ
typedef unsigned char uchar; // 8 bitovy unsigned typ

// enum types

typedef enum {//uroven ochany prikazu do konzole
 setting,
 server_setting,
 critical_setting,
 cheat,
} command_protection;
typedef enum {// konstanty pohybu
 move_fwd,move_back,
 move_up,move_down,
 move_left,move_right,
 pitch_left,pitch_right,
 turn_left,turn_right,
 turn_up,turn_down,
 use_shoot,
 nextweap,prevweap,
 scoreboard,
} move_type;
#define move_first move_fwd	//prvni z konstant pohybu
#define move_last scoreboard	//posledni z konstant pohybu
typedef enum {// konstanty klaves
 k_nothing=0,//k_nothing musi byt prvni a musi byt nula
 k_left,k_right,k_pleft,k_pright,k_up,k_down,
 k_move_fwd,k_move_back,k_move_left,k_move_right,k_move_up,k_move_down,
 k_console,k_shoot,k_nextweap,k_prevweap,
 k_score,
 k_max //k_max je specialni a musi byt za poslednim - je to pocet bindu
} keyfunc;

//3D svet

typedef struct {     //bounding box objektu
 vector org;               // pocatek
 vector dx,dy,dz;          // smery tri hran z pocatku
} bbox;
typedef struct {  // axis-aligned bounding box objektu
 vector lo;               // dolni roh (mensi souradnice)
 vector hi;               // horni roh (vetsi souradnice)
} aa_bbox;
typedef struct {//per-vertex-in-face data v modelu
 bfloat u,v; //souradnice textury ve vertexu
 int n;     //cislo vertexu co se pouzije
} facecoord;

typedef enum { //ruzne funkce objektu co se volaji
 ai_timer,
 ai_touch,
 ai_create,
 ai_damage,
 ai_destroy,
 ai_max,//posledni, urcuje pocet
} ai_func_type;

typedef struct _client client;// skutecna deklarace o kus dal

//TEXTURY

typedef struct {//hlavicka TGA souboru.
 char idbytes;     //0 je spravne, jinak-problem (na miste pixelu jsou navic jina odpadni data -> problem uzivatele)
 char colormap;    //lze ignorovat
 char type;	    //pouze 2 je spravne (RGB)
 char colmap[5];   //lze ignorovat
 short orgx,orgy;  //pocatek obrazku, ignouje se
 short x,y;        //rozmery obrazku
 char bpp;         //bits per pixel
 char flags;       //nezajimava polozka
} tgaheader;

typedef struct { //TGA soubor. Hlavicka + data
 tgaheader head;	//hlavicka tga souboru
 unsigned char data[1];	    //x*y*(bpp/8) bajtu
} tgafile;

typedef struct { // textura - pro texture manager
 int refcount;//kolikrat je pouzita. Pokud nula, lze ji odstranit
 uint id; //OpenGL handle textury
 tgafile *mem; //kde je v pameti
 char *name;//jeji jmeno
} texture;

typedef texture *ptexture;//ukazatel na texturu

//MENU

typedef struct {//polozka v menu
 char *label;	//popis polozky menu
 char *command; //prijkaz vyvolany pri zvoleni poilozky
} menuitem;
typedef struct _menusystem {
 menuitem *menu;      //pole s polozkami v menu, NULL terminated
 char *id;            //nazev menu
 struct _menusystem *next; //odkaz na dalsi prvek spojaku
} menusystem;

//CONSOLE

typedef struct _console_history { //oboustranny spojak - historie prikazove raqdky v konzoli
 char *cmd;		//prikaz
 struct _console_history *next,*prev;	//oboustranny spojak - odkaz na naslednika a predchudce
} console_history;
typedef struct { //console
 int active;				//1 je-li konzole aktivni a pise se na ni
 char cmd[console_max_cmd];		//aktualni prave psany prikaz
 int lines[console_lines],len[console_lines];	//kruhovy buffer s indexy zacatku a delkami radek v bufferu buf
 int line0,line9;			//prvni a posledni pouzita radka v linebuffer
 char buf[console_avg_width*console_lines];	//kruhovy bufer na znaky v console scrollback bufferu
 int chars,nlines;			//pocet znaku a radek v scrollback bufferu
} console_data;
typedef enum {
 float_func,
 float_var,
 int_func,
 int_var,
 string_func,
 string_var,
 vector_func,
 vector_var,
} cmd_func;

// callback funkce
typedef void vp_fn(void *v);//bere void*, nevraci nic
//typedef void *string_fn(char *s);
typedef void cmd_fn(const char *s);//prikaz - bere string, nevraci nic

typedef struct _cmd_def {
 const char *name;           //jmeno funkce/promenne
 const char *help;           //napoveda k prikazu (helptext)
 void *func;                 //adresa funkce ci promenne
 cmd_fn *special_func;       //adresa specianli funkce volane pri nastaveni promenne
 cmd_func type;		     //typ promenne (float,int,string) a funkce/promenna
 bfloat min,max;	     //minimalniu a maximalni hodnota promenne
 command_protection level;   //uroven ochrany - cheaty nelze zadavat v multiplayeru a ne vsichni mohou menit nastaveni serveru .... 
 //TODO: command protection zatim neni implementovano
 int is_saved;               //jestli se na konci ulozi do konfigurace
 struct _cmd_def *next;      //dalsi v reztezu prikazu
 //TODO: na cmd_def y mel byt BVS strom ....
} cmd_def;

//KEYBINDING

typedef struct _binding { //spojak s keybindingy
 int key;//klavesa
 int action;//bud jednoducha akce, pokud je -1, tak se spusti command
 char *command;	//vyvolany prikaz
 struct _binding *next;	//dalsi binding
} binding;

//obecne struktury

typedef struct _properties { //linked list properties, key=value,key2=value2....
 char *k; //key
 char *v; //value
 struct _properties *next; //odkaz na dalsi prvek spojaku
} properties;
//TODO: mozna predelat na binarni strom, ale jelikoz properties je vetsinou dost malo, tak to skoro ani nema smysl

typedef struct _tex_list { //spojak textur
 texture *t;               //textura
 struct _tex_list *next;    //linked list - ukazatel na dalsi. TODO: bylo by lepsi to mit stromem  
} tex_list;

typedef struct _l_list { //spojak
 void *t;              //cosi
 struct _l_list *next; //linked list - ukazatel na dalsi. TODO: nekde by bylo lepsi pouzit strom
} l_list;

/** zatim nevyresena struktura na ukladani rotace objektu. Opengl pouziva uhel + vektor */
typedef struct {
 bfloat x,y,z;//left/right,up/down,pitch, vse ve stupnich
 //todo: neslo by ten format zlepsit?
} rotation;

typedef struct { // vlastnosti objektu
 int is_physical;  //1 pokud je fyzicky a aplikuje se na nej fyzika, 2 pokud je castecne fyzicky a na nej je aplikovana fyzika, ale ostatnim v kolizi nebrani
 bfloat modelscale;// zvetseni modelu. Pokud je 0, bere se jako 1
 int modelalpha;   // alpha modelu (1-255)= alpha, 0 = vypnuto
 bfloat bounce;       //bounce prevzate z class
 bfloat health;    //health objektu
 int is_destroyed; // 1 pokud se objekt nachazi ve svem destruktoru, 2 pokud uz byl znicen, 0 jinak
 int copy_d;       // 1 pokud se ma na konci fyziky zkopirovat do d o kolik se to skutecne pohnulo
} object_features;

typedef struct {//docasna data fyzikalniho modelu pro objekt.
 bfloat touch; //0=model ve vzduch, 1= plnou silou gravitace na zemi, neco mezi = castecne na zemi(na svahu ...)
 vector movemask;//maska pohybu (co se bere z d a co povinne z nd
 bfloat fade; // konstanta zpomalovani pohybu
 int lifetime; //jak dlouho uz objekt existuje (milsekund);
} temp_physics;

typedef struct {// Struktura pro praci s textem. Pouzivana hlavne pri nahravani textovych souboru (jako napr. .map nebo .def)
 char *begin;//zacatek
 char *curr; //aktualni pozice
 int left;   //kolik bajtu zbyva;
} text_token;


#endif // of __B_TYPES_H__
