/*** prace s objekty/classami/fyzika/atd ... */
#include "b_world.h"
#include "b_net.h"

bfloat gravity=0.3;            //gravitace [EXT]
vector down={0,0,-1};		//vektor "dolu" - kam pusobi standardne gravitace [EXT]

int last_tick;//kdy byla naposledy zavolana tick funkce (v milisekundach)
int now_tick;//kdy byla naposledy zavolana tick funkce (v milisekundach)

int groundcontrol=1;	        //fyzikalni model pohybu (1=quake-like, 0=6 DOF) [EXT]

extern bfloat clipfront;
extern bfloat clipback;
extern bfloat fog_distance;
extern vector fog_color;

int nplayers;//pocet hracu [EXT]
players *plr;//spojak s hraci [EXT]

/** prida jeden objekt do seznamu objektu a inicializuje ho. Je-li mala kapacita, zvetsi ji. Vrati ukazatel na pridany objekt */
object3d *world_add_empty_object(worldspawn *w) {
 int i;
 object3d *last;
 if (w->obj.numobj>=w->obj.alocobj) {//malo objektu, nutno prialokovat
  w->obj.obj=realloc1(w->obj.obj,w->obj.alocobj*2*sizeof(void*));
  for (i=w->obj.alocobj;i<w->obj.alocobj*2;i++) world_getobj(w,i)=0;//vynulovani toho zbytku pointeru
  w->obj.alocobj*=2;
 }
 world_lastobj(w)=last=malloc0(sizeof(object3d));
 w->obj.numobj++;
 last->id=(int)last;
 last->lastupdate=gettickcount();
 net_addobject(last);
 return last;
}

/** vytvori novy objekt, umisteny pocatkem v objektu src */
object3d *world_add_object_from(worldspawn *w,object3d *src) {
 object3d *newobj=world_add_empty_object(w);
 if (src) { //zkopiruje do hrace udaje (pozice, orientace....
  newobj->head=src->head;
  newobj->left=src->left;
  newobj->up=src->up;
  newobj->rot=src->rot;
  newobj->rotplus=src->rotplus;
  newobj->p=src->p;
 }
 return newobj;
}

/** odstrani ze sveta jeden objekt */
void world_kill_object(worldspawn *w,object3d *who) {
 int i,j;
 net_killobject(who);
 if (!w->obj.numobj) fatalf("world_kill_object: World empty - cannot kill");
 world_foreach_obj(w,i) {
  if (world_getobj(w,i)==who) {// objekt nalezen, bude znicen
   for (j=i;j<w->obj.numobj-1;j++) world_getobj(w,j)=world_getobj(w,j+1);
   w->obj.numobj--;
   freeobject(who);
   return;
  }
 }
 fatalf("world_kill_object: cannot kill: object %d not found",who);
}

tree *class_tree; // strom s nahranymi classami [EXT]

const ai_func_type parse_ai[ai_max]={ ai_timer, ai_touch, ai_create, ai_damage, ai_destroy};
const char parse_ai2[ai_max][11]={ "ai_timer", "ai_touch", "ai_create", "ai_damage","ai_destroy"};

/** z properties ziska vlastnost jako string */
char *getprop(properties *p,const char *name) {
 while (p) {
  if (eqi(p->k,name)) return p->v;
  p=p->next;
 }
 debug_printf0("getprop: Property %s not found\n",name);
 return 0;//nothing found
}

/** z properties ziska odkaz na string (vhodne kdyz ho chceme zmenit)*/
char **getprop_a(properties *p,const char *name) {
 while (p) {
  if (eqi(p->k,name)) return &p->v;
  p=p->next;
 }
 debug_printf0("getprop_a: Property %s not found\n",name);
 return 0;//nothing found
}

/** z objektu ziska jeho vlastnost jako int */
int object_getpropi(object3d *o,const char *name) {
 return atoi(unnull(getprop(o->prop,name)));
}

/** z objektu ziska jeho vlastnost jako float */
bfloat object_getpropf(object3d *o,const char *name) {
 return atof(unnull(getprop(o->prop,name)));
}

/** z objektu ziska jeho vlastnost jako float, pokud neexistuje, vrati defvalue */
bfloat object_getpropfd(object3d *o,const char *name,bfloat defvalue) {
 char *v=getprop(o->prop,name);
 if (v) return atof(v);
 return defvalue;
}

/** z objektu ziska jeho vlastnost jako int, pokud neexistuje, vrati defvalue */
int object_getpropid(object3d *o,const char *name,int defvalue) {
 char *v=getprop(o->prop,name);
 if (v) return atoi(v);
 return defvalue;
}

/** z objektu ziska jeho vlastnost jako float, vrati 0 pokud neexistuje */
int object_getpropfv(object3d *o,const char *name,bfloat *ret) {
 char *x;
 if (!o) return 0;//objekt neexistuje
 x=getprop(o->prop,name);
 if (!x) return 0;
 *ret=atof(x);
 return 1;
}

/** prida key=value do properties, zkontroluje jestli tam neni */
void property_add(properties **p,const char *key,const char *value) {
 properties *pp;
 char **v=getprop_a(*p,key);
 if (!v) {//neni tam
  pp=malloc1(sizeof(properties));
  pp->k=strdup(key);
  pp->v=strdup(value);
  pp->next=*p;
  *p=pp;
 } else {//dany klic tam uz je 
  free(*v);
  *v=strdup(value);
 }
}

/** zkopiruje properties z classy na konkretni objekt, nektere specialni nastavi */
void properties_copy_class_to_object(objectclass *c,object3d *o) {
 properties *p=c->prop;
 while (p) { 
  if (eqi(p->k,"bounce")) {
   o->f.bounce=atof(p->v);
  } else if (eqi(p->k,"model_rot_x")) {
   o->rot.x=atof(p->v);
  } else if (eqi(p->k,"model_rot_y")) {
   o->rot.y=atof(p->v);
  } else if (eqi(p->k,"model_rot_z")) {
   o->rot.z=atof(p->v);
  } else {
   property_add(&o->prop,p->k,p->v);
  }
  p=p->next;
 }
}

//TODO: properties jako tree

#ifndef only_console //[PASS]

/** z retezce vyparsuje axis-aligned bounding box */
aa_bbox *parse_abox(const char *def) {
 float i[6];//bbox
 int j;
 aa_bbox *abox;
 char *value;
 if (!def) return 0;//null string
 if (!def[0]) return 0;//empty string
 value=strdup(def);
 str_replace(value,',',' ');
 j=sscanf(value,"%f %f %f %f %f %f",i+0,i+1,i+2,i+3,i+4,i+5);
 free(value);
 if (j!=6 && j!=3) {
  return 0;
 } else {
  if (j==3) {//kratsi symetricky bbox
   for (j=0;j<3;j++) {
    i[j]/=-2;
    i[j+3]=-i[j];
   }
  }
  abox=malloc1(sizeof(aa_bbox));
  abox_inf(abox);
  abox_addvertex(abox,vect(i[0],i[1],i[2]));
  abox_addvertex(abox,vect(i[3],i[4],i[5]));
  return abox;
 }
}


/** z classy, do ktere byla nahrana jenom definice nahraje podle te definice i zbytek. Definici pote smazne*/
void parse_classdef(objectclass *oc) {
 text_token t;
 int j;
 char *key,*value;
#ifdef old_code
 float i[3]={1,1,1};//bbox
#else
 float i[6];//bbox
#endif
 t.curr=t.begin=oc->definition;//init token
 t.left=strlen(t.begin);
 debug_printf0("[ClassMan] parsing definition: %s\n",oc->classname);
 oc->prop=0;
 while (token_readkv(&t,&key,&value)) {//foreach key & value in definition
  debug_printf0("%s: %s = %s\n",oc->classname,key,value);
#ifdef old_code
  if (eqi(key,"modelscale")) {
   str_replace(value,',',' ');
   if (!sscanf(value,"%f %f %f",&i[0],&i[1],&i[2])==3) {
    i[0]=i[1]=i[2]=1;
    bug_report("Invalid vector"); 
   }
  } else if (eqi(key,"model")) {
   oc->model=b_loadmodel_r_trans(value,0,0,0,i[0],i[1],i[2]);
#else
  if (eqi(key,"bbox")) {
   oc->abox=parse_abox(value);
   if (!oc->abox) bug_report("Invalid two vectors"); 

/*   str_replace(value,',',' ');
   j=sscanf(value,"%f %f %f %f %f %f",i+0,i+1,i+2,i+3,i+4,i+5);
   if (j!=6 && j!=3) {
    bug_report("Invalid two vectors"); 
   } else {
    if (j==3) {//kratsi symetricky bbox
     for (j=0;j<3;j++) {
      i[j]/=-2;
      i[j+3]=-i[j];
     }
    }
    oc->abox=malloc1(sizeof(aa_bbox));
    abox_inf(oc->abox);
    abox_addvertex(oc->abox,vect(i[0],i[1],i[2]));
    abox_addvertex(oc->abox,vect(i[3],i[4],i[5]));
   }*/
  } else if (eqi(key,"modelscale")) {
   str_replace(value,',',' ');
   if (!sscanf(value,"%f %f %f",&i[0],&i[1],&i[2])==3) {
    i[0]=i[1]=i[2]=1;
    bug_report("Invalid vector"); 
   }
   for (j=0;j<3;j++) {//0 = default = 100% velikosti
    if (!i[j]) i[j]=1;
   }
   v(oc->modelscale,i[0],i[1],i[2]);
  } else if (eqi(key,"model")) {
#endif
   oc->model=b_loadmodel_r(value);
  } else if (eqi(key,"physical")) {
   oc->is_physical=atoi(value);
  } else if (eqi(key,"health")) {
   oc->health=atof(value);
  } else { // it is ai_* or nothing ...
   for (j=0;j<=ai_max;j++) {
    if (j==ai_max) {//neco neznameho, tak je to proste property
     property_add(&oc->prop,key,value);
     //bug_report("Unknown property");
     break;
    }
    if (eqi(key,parse_ai2[j])) {
     ai_assign(oc,value,parse_ai[j]);
     break;
    }
   }
  }
  free(key);
  //value se nesmi dealokovat
 }
 free(oc->definition);
 oc->definition=0;
}
#endif //[PASS]

char *static_cat_buf;//staticky buffer pro funkce static_cat*
int static_cat_buf_size;//velikost statickeho bufferu pro funkce static_cat*

/** vrati odkaz na buffer, obsahujici tri spojene retezce.
 Buffer je "staticky" a pri pristim volani funkce bude zmenen*/
char *static_cat3(const char *s1,const char *s2,const char *s3) {
 int l=strlen(s1)+strlen(s2)+strlen(s3);
 if (static_cat_buf_size<l+1) {
  if (static_cat_buf) free(static_cat_buf);
  static_cat_buf=malloc1(l+1);
  static_cat_buf_size=l+1;
  debug_printf0("static_cat3: buffer reallocated to %d\n",l+1);
 }
 strcpy(static_cat_buf,s1);
 strcat(static_cat_buf,s2);
 strcat(static_cat_buf,s3);
 return static_cat_buf;
}

/** z daneho classname vrati odkaz na patricny typ (pokud typ neexistuje, tak je to celkem fatalni chyba)
 need_parse: ma-li se parsovat (1), nebo staci jen ujistit, ze existuje (0) (napr. pro mapload, atd ...) */
objectclass *b_loadclass(const char *name,int need_parse) {
 objectclass *oc;
 char *def;
 int defsize;
 if (!name || !*name) return 0;//nekdo chtel classu s prazdnym jmenem : dostane NULL :o)
 oc=tree_search(class_tree,name);
 if (oc) {
  #ifndef only_console
  if (oc->definition && need_parse) parse_classdef(oc);
  #else 
  unused(need_parse);
  #endif
  debug_printf0("[ClassMan] %s used directly\n",name);
  oc->refcount++;
  return oc; //jiz v pameti, netreba nahravat
 }
 debug_printf("[ClassMan] %s not exist, loading...\n",name);
 defsize=fs_loadfile((void **)(&def),static_cat3("classes/",name,".def"));
 if (!def) {
  error_printf("Cannot load class %s",name);
  fatal_loading_error=1;
  return 0;
 }
 oc=malloc0(sizeof(objectclass));//vytvorime classu
 oc->classname=strdup(name);
 oc->refcount=1; //todo: refcount (nekde se musi zas ubirat!)
 def=realloc(def,defsize+1);
 def[defsize]=0;//null terminated string
 oc->definition=def;
 #ifndef only_console
 if (need_parse) parse_classdef(oc);
 #endif
 //TODO; dodelat nejaky loader, co bude ty classy cpat do stromu (nahraje vse z base.def) !!!!!!!!!!!!!!!!!!
 tree_add(&class_tree,name,oc);//pridame do pameti
 return oc;
}

/** najde pocet objektu daneho classname ve svete */
int findobjectnum(worldspawn *w,const char *classname) {
 int i,num=0; 
 object3d *o;
 debug_printf0("findobjectnum (%s)\n",classname);
 if (!w->obj.numobj) return 0; //vytvori defaultniho hrace
 for (i=0;i<w->obj.numobj;i++) {
  o=world_getobj(w,i);
  if (!o) fatalf("nullobject in world: %d",i);
  if (!o->type) continue;
  debug_printf0("trying %s\n",o->type->classname);
  if (eqi(o->type->classname,classname)) num++;
 }
 return num;
}

/** najde objekt daneho classname ve svete (poradi num) . Poradi je 0..NUM_OBJECTS-1 */
object3d *findobject(worldspawn *w,const char *classname,int num) {
 int i; 
 object3d *o;
 debug_printf0("findobject (%d of %s)\n",num,classname);
 if (!w->obj.numobj) return 0; //vytvori defaultniho hrace
 for (i=0;i<w->obj.numobj;i++) {
  o=world_getobj(w,i);
  if (!o->type) continue;
  debug_printf0("trying %s\n",o->type->classname);
  if (eqi(o->type->classname,classname)) {
   if (num<=0) return o;
   debug_printf0("%d too early\n",num);
   num--;
  }
 }
 debug_printf("findobject: %s not found\n",classname);
 return 0;
}

/** najde objekt daneho classname ve svete (je-li jich vic, jeden nahodne vybere) */
object3d *findobject_random(worldspawn *w,const char *classname) {
 debug_printf0("findrand %s\n",classname);
 return findobject(w,classname,brandom(findobjectnum(w,classname)));
}

/** zastavi worldspawn a odpoji hrace */
void world_stop(worldspawn *w) {
 #ifdef have_sound
 b3d_stop_sounds();
 #endif
 world_collision_free(w);
 player_free_all();
 unused(w);
}

#pragma warning(disable : 4706) // assignment within conditional expression
/** nastartuje worldspawn
 (a vytvori hrace pokud tam neexistuje) */
void world_start(worldspawn *w) {
 client *player1;
 char *m;
 bfloat old;
 intermission_end();//skonci intermission, pokud byla
 player1=addhumanplayer(w,0);
 //TODO: nez zacnu ty veci vyndavat, tak nastavit defaultni nastaveni: zadny skybox, rozumny fog/clip
 if (m=getprop(w->prop,"message")) con_print(m);
 if (m=getprop(w->prop,"fogdist")) {
  fog_distance=atof(m);
  if (fog_distance<0) {
   debug_printf("Bad fog distance: %d\n");
   fog_distance=0;
  }
 }
 if (m=getprop(w->prop,"fogcolor")) fog_color=atov(m);
 if (m=getprop(w->prop,"farclip")) {
  old=clipback;
  clipback=atof(m);
  if (clipback<clipfront) {  
   debug_printf("Bad clipback: %d\n");
   clipback=old;
  }
 }
 if (m=getprop(w->prop,"skyname")) fs_loadskymap(m);
 world_collision_init(w);
 last_tick=now_tick=gettickcount();//vyresetovani citace
 //TODO: init funkce objektu by se mozna mela volat tady a ne uz pri nahravani
}
#pragma warning(default : 4706) // assignment within conditional expression

int screens;//pocet obrazovek ktere jsou obsazene hracem ci kamerou [EXT]
int screens_max=default_screens_max;//maximalni pocet obrazovek ktere jsou obsazene hracem ci kamerou [EXT]
int player_screens;//pocet obrazovek ktere jsou obsazene hrace, ktereho lze ovladat [EXT]

client **vplayer; //viditelni hraci (na obrazovce). Pole ukazatelu [EXT]
int activeplayer=1;//ktery hrac/kamera/atd ... je prave aktivni (ma focus) [EXT]
client *mouseplayer;//hrac ktery ma mys (je ovladan mysi, jakozto jediny) [EXT]

/** vytvori prazdneho, neinicializovaneho hrace pomoci objektu */
client *player_add_empty(object3d *o) {
 client *c;
 players **p;
 //vytvori prazdneho hrace
 c=malloc0(sizeof(client));
 c->who=o;
 //prida hrace do seznamui
 p=&plr;
 while (*p) p=&((*p)->next);
 *p=malloc0(sizeof(players));
 (*p)->next=0;
 (*p)->p=c;
 nplayers++;
 return c;
}

/** zlikviduje vsechny hrace (a take kamery a podobne objekty ...) */
void player_free_all() {
 players *p;
 while (plr) {
  p=plr->next;
  free(plr->p);
  plr=p;
 }
 nplayers=0;
 screens=0;
 player_screens=0;
}

/** prida do sveta jednoho hrace, vzaneho ze startu.
 Pokud client neni null, bude hrac prirazen tomu klientovi, jinak bude vytvoren klient novy */
client *addhumanplayer(worldspawn *w,client *player1) {
 bfloat rot=0;
 object3d *player,*playerx;
 player=findobject_random(w,"info_player_start");
 debug_printf0("findplayer %d\n",player);
 playerx=world_add_object_from(w,player);
 rot=0;
 if (!player || object_getpropfv(player,"angle",&rot)) { //nastaveni uhle u hrace
  rot=(90.0-rot)*pi/180.0;//prepocet z quakovskeho uhle
  //uhel pro full 3d pohled
  v(playerx->head,sin(rot),cos(rot),0);
  v(playerx->up,0,0,1);
  v_mulv2(&playerx->left,playerx->head,playerx->up);
 }
 class_set(playerx,"player");
 //uhel pro quake-like pohled
 if (player1) { //hrac existuje (jen se respawnuje)
  playerx->client=player1;
  player1->who=playerx;
 } else player1=player_add_human(playerx);
 player1->data.human.r_angle=0;
 player1->data.human.r_angle2=0;
 player1->data.human.r_angle3=0;
 if (groundcontrol) player1->data.human.r_angle=rot;
 world_updateobject(w,playerx);
 return player1;
}

/** vytvori hrace (rizeneho clovekem) */
client *player_add_human(object3d *o) {
 client *c=player_add_empty(o);
 c->controller=human;
 o->client=c;
 player_allocate_screen(c);
 return c;
}

/** Pro daneho hrace naalokuje jednu obrazovku (vytvori tak split screen kdyz je treba) */
void player_allocate_screen(client *c) {
 if (screens==screens_max) {
  screens_max*=2;
  vplayer=realloc1(vplayer,screens_max*sizeof(client*));
 }
 vplayer[screens]=c;
 activeplayer_check();
 c->vis=screen;
 c->data.human.control=player_screens;//TODO: muzou tam byt diry po mazani, je nutne porjit ten seznam!
 screens++;
 player_screens++;
 player_realloc_screens();
}

/** prehazi hrace na obrazovce, aby ji zabirali rovnomerne (priblizne) */
void player_realloc_screens() {
 int i,j,t;
 int nx,ny;//pocet obravoek ve rozmerech X a Y
 int left;//pocet obrazovek k rozdeleni
 if (!screens) return;
 if (screens==1) {
  vplayer[0]->x1=0;
  vplayer[0]->y1=0;
  vplayer[0]->x2=maxx;
  vplayer[0]->y2=maxy;
  return;
 }
 //rovnomerny splitscreen
 left=screens;
 ny=msqrt(screens);
 t=0;
 for (i=0;i<ny;i++) {
  nx=(left+ny-i-1)/(ny-i);
  for (j=0;j<nx;j++) {
   vplayer[t]->x1=(maxx/nx)*j;
   vplayer[t]->y1=(maxy/ny)*i;
   vplayer[t]->x2=(maxx/nx)*(j+1);
   vplayer[t]->y2=(maxy/ny)*(i+1);
//   con_printf("%d: %d %d - %d %d",t,vplayer[t]->x1,vplayer[t]->y1,vplayer[t]->x2,vplayer[t]->y2);
   t++;left--;
  }
 }
}

/** zkontroluje jestli je nastaven dobre akvivni hrac a pripadne poopravi. Take nastavi mysi focus */
void activeplayer_check() {
 debug_printf0("player may switch to: %d - %d\n",1,nplayers);
 if (activeplayer<1) activeplayer=nplayers;
 if (activeplayer>nplayers) activeplayer=1;
 debug_printf0("player switch to: %d\n",activeplayer);
 mouseplayer=client_getn(activeplayer);
}

/** vrati hrace s cislem n (pocitano od 1) */
client *client_getn(int n) {
 players *p=player_getn(n);
 if (p) return p->p;
 return 0;
}

/** vrati hrace s cislem n (pocitano od 1) */
players *player_getn(int n) {
 players *pl=plr;
 if (!pl) return 0;
 while (pl) {
  if (n<=1) return pl;
  n--;
  pl=pl->next;
 }
 return 0;
}

/** vrati pocet hracu */
int player_getnum(void) {
 players *pl=plr;
 int n=0;
 while (pl) {
  n++;
  pl=pl->next;
 }
 return n;
}

