/*** Transparency fix - nastavi spravnou transparenci (alpha) u fontu */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#define stat _stat	//ve windows se jmenuje funkce jinak
#endif

#ifndef max
#define max(x,y) (((x)>(y))?(x):(y)) //maximum ze dvou cisel
#endif

typedef struct {//hlavicka TGA souboru.
 char idbytes;     //0 je spravne, jinak-problem (na miste pixelu jsou navic jina odpadni data -> problem uzivatele)
 char colormap;    //lze ignorovat
 char type;	    //pouze 2 je spravne (RGB)
 char colmap[5];   //lze ignorovat
 short orgx,orgy;  //pocatek obrazku, ignouje se
 short x,y;        //rozmery obrazku
 char bpp;         //bits per pixel
 char flags;       //nezajimava polozka
} tgaheader;

tgaheader th;//nacteny targa header

/** vrati velikost souboru */
int fsize(char *f) {
 struct stat b;
 if (stat(f,&b)) return 0;//return 0 if error
 return b.st_size;
}

typedef struct {//pixel - rgb a alpha komponenty
 unsigned char a[3];//RGB
 unsigned char t;//alpha
} q;

/** vypise chybovou hlasku a ukonci program */
void fatal(const char *s) {
 printf("ERROR: %s\n",s);
 exit(-1);
}

/** main of the transparency utility */
int main(int argc,char **argv) {
 FILE *f,*g;
 int s,i,j,maxj,show=0;
 char mode='f';//font
 q *a;
 printf("Usage: trans <in.tga> <out.tga> [<mode>] [<doRGB>]\nMode can be f (font, default) or e (explosion)\nif doRGB specified, mask is stored also in RGB");
 if (argc<3) exit(-1);
 if (argc>=4) mode=tolower(argv[3][0]);
 if (argc>=5) show=1;
 f=fopen(argv[1],"rb");
 s=fsize(argv[1]);
 if (!f || s<18+4) fatal("Cannot read input file");
 fread(&th,18,1,f);
 if (th.type!=2) fatal("TGA must be uncompressed RGB");
 if (th.bpp!=32) fatal("TGA must be 32bit");
 if (th.idbytes) fatal("TGA must not contain any extra data");
 g=fopen(argv[2],"wb");
 if (!g) exit(-1);
 a=malloc(s);
 fwrite(&th,18,1,g);
 fread(a,s-18,1,f);
 printf("%d pixels\n",(s-18)/4);
 if (mode=='e') {//explosion mode
  maxj=0;
  for (i=0;i<(s-18)/4;i++) {//for each pixel
   j=a[i].a[0];
   j+=a[i].a[1];
   j+=a[i].a[2];//j now contains intensity value
   maxj=max(maxj,j);
  }
  printf("Maximum intensity: %d\n",maxj);
//  maxj/=2;
 }
 for (i=0;i<(s-18)/4;i++) {//for each pixel
  if (mode=='e') {//explosion mode
   j=a[i].a[0];
   j+=a[i].a[1];
   j+=a[i].a[2];//j now contains intensity value
   a[i].t=j*255/maxj;
   if ((j*255)/maxj>255) a[i].t=255;
  } else {//font mode
   j=a[i].a[0];
   j+=a[i].a[1]*256;
   j+=a[i].a[2]*256*256;//j now contains RGB value
   if (j==0xff || j==0xff0000) {//100% red or 100% blue
    j=0;//..is converted to 100% black
    a[i].a[0]=0;
    a[i].a[1]=0;
    a[i].a[2]=0;
   }
   if (j) {//if not black
    a[i].t=255;//opaque
   } else {// black
    a[i].t=0;//transparent
   }
  }//end of mode
  if (show) {
   a[i].a[0]=a[i].a[1]=a[i].a[2]=a[i].t;
  }
 }
 fwrite(a,s-18,1,g);
 free(a);
 fclose(f);
 fclose(g);
 printf("%s -> %s\n",argv[1],argv[2]);
 return 0;
}